/**************************************************************
Module : BankAcct.cpp - a multithreaded application sample
             Written 2000 by Dmitri Leman
Purpose: To demonstrate ability of tracing framework to collect
    traces from multiple threads
**************************************************************/
#ifdef WIN32
#include <windows.h>
#else
#include <unistd.h>
#include <pthread.h>
#include <sys/time.h>
#include <sys/sem.h>
#include <errno.h>
#include <string.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "HTrace.h"

const char s_szNumThreadsKeyWord[] = "-NumThreads=";
const char s_szPriorityKeyWord[] = "-Priority=";
const char s_szAffinityKeyWord[] = "-Affinity=";

#ifdef WIN32
static HANDLE s_hEvent = NULL;
#else
static int s_iSemID = 0;
int GetCurrentThreadId()
{
    return (int)getpid();
}
#endif
static bool   s_bStopThreads = false;

static int    s_iAccount = 1000;

/*------------------------------------------------------------

   FUNCTION: ThreadProc

   PURPOSE: Thread procedure for manipulation on the account.
      
-------------------------------------------------------------*/
#ifdef WIN32
DWORD WINAPI ThreadProc(LPVOID lpParameter)
#else
void * ThreadProc(void * lpParameter)
#endif
{
  //Wait until all threads are created and the event is raised
    #ifdef WIN32
        WaitForSingleObject(s_hEvent, 5000);
    #else
	sembuf l_SemBuf;
	l_SemBuf.sem_num = 0;
	l_SemBuf.sem_op = -1;
	l_SemBuf.sem_flg = SEM_UNDO;
	if(semop(s_iSemID, &l_SemBuf, 1) != 0)
	{
	    HTRACE(TG_Error, "semop failed. %s\n",strerror(errno));
	    return 0;
	}
    #endif
    printf("BankAcct: Thread %d started\n", 
        GetCurrentThreadId());
    static long ls_lErrorDetected = -1;
    while(!s_bStopThreads)
    {
        int l_iAmount = 100;
       //This is an intentional bug: no synchronization is done
        //while reading, comparing and modifying the account.

        if(ls_lErrorDetected < 0)
        {
            HTRACE(TG_BankAccount, 
                "BankAcct: Thrd %x Acc %d check account", 
                GetCurrentThreadId(), s_iAccount);
        }

        if(s_iAccount >= l_iAmount)
        {
            //Create a delay to increase chance of 
            //account corruption
            for(volatile int j = 0; j < 100; j++);

            s_iAccount -= l_iAmount;

            if(ls_lErrorDetected < 0)
            {
                HTRACE(TG_BankAccount, 
                    "BankAcct: Thrd %x Acc %d after withdraw", 
                    GetCurrentThreadId(), s_iAccount);
            }

            if(s_iAccount < 0)
            {
                if(InterlockedIncrement(&ls_lErrorDetected) >0)
                {
                    InterlockedDecrement(&ls_lErrorDetected);
                    continue;
                }

                HTRACE(TG_BankAccount, 
                    "BankAcct: Error in the Bank Account: "
                    "Thread %x Amount = %d", 
                     GetCurrentThreadId(), s_iAccount);
                printf(
                    "Thread %d. Account %d. Bank account < 0. "
                    "Do you want to deposit or terminate the "
                    "app? Click 'q' to terminate\n",
                    GetCurrentThreadId(), s_iAccount);
                if(getchar() == 'q')
                {
                    InterlockedDecrement(&ls_lErrorDetected);
                    s_bStopThreads = true;
                    printf("Thread %d ended\n", 
                        GetCurrentThreadId());
                    return 0;
                }
                InterlockedDecrement(&ls_lErrorDetected);
                s_iAccount = 500;
            }
        }//if(s_iAccount >= l_iAmount)
        else 
            s_iAccount += l_iAmount;
        
        if(ls_lErrorDetected < 0)
        {
            HTRACE(TG_BankAccount, "BankAcct: Thrd %x Acc %d", 
                GetCurrentThreadId(), s_iAccount);
        }
        
    }
    printf("Thread %d ended\n", GetCurrentThreadId());
    return 0;
}//DWORD WINAPI ThreadProc

/*------------------------------------------------------------

   FUNCTION: LaunchThreads

   PURPOSE: Read command line and create threads
      
-------------------------------------------------------------*/
#ifdef WIN32
BOOL LaunchThreads
(
    int p_iNumThreads,
    int p_iPriority,
    int p_iAffinity
)
{
    int l_iPriorities[][3] = 
    {
    { 1, IDLE_PRIORITY_CLASS, THREAD_PRIORITY_IDLE           },
    { 2, IDLE_PRIORITY_CLASS, THREAD_PRIORITY_LOWEST         },
    { 3, IDLE_PRIORITY_CLASS, THREAD_PRIORITY_BELOW_NORMAL   },
    { 4, IDLE_PRIORITY_CLASS, THREAD_PRIORITY_NORMAL         },
    { 5, NORMAL_PRIORITY_CLASS, THREAD_PRIORITY_LOWEST       },
    { 6, NORMAL_PRIORITY_CLASS, THREAD_PRIORITY_BELOW_NORMAL },
    { 7, NORMAL_PRIORITY_CLASS, THREAD_PRIORITY_NORMAL       },
    { 8, NORMAL_PRIORITY_CLASS, THREAD_PRIORITY_ABOVE_NORMAL },
    { 9, NORMAL_PRIORITY_CLASS, THREAD_PRIORITY_HIGHEST      },
    {11, HIGH_PRIORITY_CLASS, THREAD_PRIORITY_LOWEST         },
    {12, HIGH_PRIORITY_CLASS, THREAD_PRIORITY_BELOW_NORMAL   },
    {13, HIGH_PRIORITY_CLASS, THREAD_PRIORITY_NORMAL         },
    {14, HIGH_PRIORITY_CLASS, THREAD_PRIORITY_ABOVE_NORMAL   },
    {15, HIGH_PRIORITY_CLASS, THREAD_PRIORITY_TIME_CRITICAL  },
    {16, REALTIME_PRIORITY_CLASS,THREAD_PRIORITY_IDLE       },
    {22, REALTIME_PRIORITY_CLASS,THREAD_PRIORITY_LOWEST     },
    {23, REALTIME_PRIORITY_CLASS,THREAD_PRIORITY_BELOW_NORMAL},
    {24, REALTIME_PRIORITY_CLASS,THREAD_PRIORITY_NORMAL      },
    {25, REALTIME_PRIORITY_CLASS,THREAD_PRIORITY_ABOVE_NORMAL},
    {26, REALTIME_PRIORITY_CLASS,THREAD_PRIORITY_HIGHEST     },
    {31, REALTIME_PRIORITY_CLASS,THREAD_PRIORITY_TIME_CRITICAL}
    };//int l_iPriorities[3][] = 
    int l_iNumPriorities = 
        sizeof(l_iPriorities)/sizeof(l_iPriorities[0]);
    bool l_bFound = false;
    int  i;
    for(i = 0; i < l_iNumPriorities; i++)
    {
        if(l_iPriorities[i][0] == p_iPriority)
        {
            l_bFound = true;
            break;
        }
    }
    if(!l_bFound)
    {
        HTRACE(TG_Error, 
            "ERROR: wrong priority %d\n", p_iPriority);
        return FALSE;
    }
    int l_iPriorityClass = l_iPriorities[i][1];
    int l_iPriorityThread = l_iPriorities[i][2];

    SetPriorityClass(GetCurrentProcess(), l_iPriorityClass);

    s_hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);

    DWORD * l_pdwThreadID = new DWORD[p_iNumThreads];
    HANDLE * l_hThread = new HANDLE[p_iNumThreads];
    int l_iStillRunning = p_iNumThreads;
    for(i = 0; i < p_iNumThreads; i++)
    {
        l_hThread[i] = CreateThread(NULL, 0, ThreadProc, NULL,
            0/*CREATE_SUSPENDED*/ , &l_pdwThreadID[i]);
        if(l_hThread[i] == NULL)
        {
            HTRACE(TG_Error,"ERROR: failed to start thread\n");
        }
        else
        {
            SetThreadPriority(l_hThread[i], l_iPriorityThread);
            SetThreadAffinityMask (l_hThread[i], p_iAffinity);
            l_iStillRunning++;
        }
    }//for(i = 0; i < l_iNumThreads; i++)

    SetEvent(s_hEvent);//start threads
    char l_szMessage[512];
    sprintf(l_szMessage, "BankAcct application started. "
     "Number of threads = %d, Priority = %d, Affinity = %x\n\n"
        "Press 'q' and Enter key to terminate the application",
        p_iNumThreads, p_iPriority, p_iAffinity);
        
    printf(l_szMessage);
    while(getchar() != 'q');
    s_bStopThreads = true;
   
    while(l_iStillRunning)
    {
        l_iStillRunning = 0;
        for(i = 0; i < p_iNumThreads; i++)
        {
            if(l_hThread[i] == NULL)
                continue;

            if(WAIT_OBJECT_0 == 
                WaitForSingleObject(l_hThread[i], INFINITE))
            {
                CloseHandle(l_hThread[i]);
                l_hThread[i] = NULL;
            }
        }//for(i = 0; i < l_iNumThreads; i++)
    }//while(l_iStillRunning)
    CloseHandle(s_hEvent);
    return TRUE;
}//BOOL LaunchThreads
#elif defined(__linux__)
bool LaunchThreads
(
    int p_iNumThreads,
    int p_iPriority,
    int p_iAffinity
)
{
    s_iSemID = semget((key_t)time(NULL), 1, IPC_CREAT);
    if(s_iSemID < 0)
    {
        HTRACE(TG_Error, "ERROR: semget failed. %s",
            strerror(errno));
        return false;
    }
    pthread_t * l_pThreads = new pthread_t[p_iNumThreads];
    int i, l_iStarted = 0;
    for(i = 0; i < p_iNumThreads; i++)
    {
        int l_iRes = pthread_create(l_pThreads+i, NULL, 
            ThreadProc, NULL);
        if(l_iRes != 0)
        {
            HTRACE(TG_Error, 
                "ERROR: pthread_create failed %s\n", 
                strerror(errno));
            break;
        }
        l_iStarted++;
    }
    
    //Increment semaphore to start all threads simultaneously
    semctl(s_iSemID, 0, SETVAL, l_iStarted);
    printf("%d threads started\nClick 'q' to terminate.", 
        l_iStarted);
    while(getchar() != 'q');
    s_bStopThreads = true;
    for(i = 0; i < l_iStarted; i++)
    {
        void * l_pThreadResult;
        int l_iRes = pthread_join(l_pThreads[i], 
            &l_pThreadResult);
        if(l_iRes != 0)
        {
            printf("ERROR: pthread_join failed %s\n", 
                strerror(errno));
        }         
    }
    return true;
}
#endif
/*------------------------------------------------------------

   FUNCTION: WinMain

   PURPOSE: Initial entry point for the application.
      
   PARAMETERS:                   .
                                  
   HINSTANCE p_hInstance            handle to current instance
   HINSTANCE p_hPrevInstance        not used
   LPSTR  p_lpszCmdLine          pointer to command line
   int    p_nCmdShow             show state of window

   RETURN VALUE:
      int                        Exit value
-------------------------------------------------------------*/
int main(int argc, char ** argv)
{
    int l_iNumThreads = -1;
    int l_iPriority = -1;
    int l_iAffinity = -1;
    int i;

    char * p;
    for(i = 1; i < argc; i++)
    {
        char * l_pParams = argv[i];
        bool l_bUnrecognized = true;
        if((p = strstr(l_pParams,s_szNumThreadsKeyWord))!=NULL)
        {
            l_bUnrecognized = false;
            l_iNumThreads = strtol
                (p+sizeof(s_szNumThreadsKeyWord)-1, NULL, 0);
        }
        if((p =strstr(l_pParams, s_szPriorityKeyWord))!=NULL)
        {
            l_bUnrecognized = false;
            l_iPriority = 
                strtol(p + sizeof(s_szPriorityKeyWord)-1,
            NULL, 0);
        }
        if((p = strstr(l_pParams, s_szAffinityKeyWord))!=NULL)
        {
            l_bUnrecognized = false;
            l_iAffinity = 
                strtol(p + sizeof(s_szAffinityKeyWord)-1,
            NULL, 0);
        }
        if(l_bUnrecognized)
        {
           printf("Parameter %s unrecognized\n"
               "Usage: BankAcct %snum %snum %snum\n",
               s_szNumThreadsKeyWord, 
               s_szPriorityKeyWord,
               s_szAffinityKeyWord);
           return 1;
        }
    }//for(i = 1; i < argc; i++)

    if(l_iNumThreads < 1)
       l_iNumThreads = 2;
    if(l_iPriority < 0)
       l_iPriority = 7;
    if(l_iAffinity < 0)
       l_iAffinity = 1;

    TraceInitialize();
    //By default, show errors and warnings on the
    //debugger's log window and in message boxes.
    TraceAssignGroupsToStream(TO_DebugMonitor | TO_MessageBox,
        TG_MessageBox | TG_Error, (DWORD)-1);
    
    //Print bank account and thread swaps(if the driver is 
    //loaded) to the memory buffer
    TraceAssignGroupsToStream(TO_MemoryBuffer,
        TG_BankAccount | TG_ThreadSwaps, (DWORD)-1);

    //Print application debug messages to the debugger's window
    TraceAssignGroupKeyWordToStream
        (TO_DebugMonitor,KeyWordAppDebug);

    #ifdef WIN32
    char l_szFullINIPath[_MAX_PATH];
    GetFullPathName("HTrace.ini", 
        sizeof(l_szFullINIPath), l_szFullINIPath, NULL);

    //Read state of all streams. This can override the 
    //assignments made in the preceding calls.
    TraceReadWriteSettingsFromIniFile(l_szFullINIPath, false,
        (DWORD)-1);
    #endif


    LaunchThreads(l_iNumThreads, l_iPriority, l_iAffinity);

    TraceDumpBufferToFile("Trace.txt");
    TraceFreeBuffer();

    TraceUnInitialize();
    return 0;
}//int APIENTRY WinMain

