//----------------------------------------------------------------------
//
//  Copyright(c) 2001, Richard Nies
//
//  Permission granted to use this code without restriction as
//  long as this copyright notice appears in all source files.
//
//  Class:        CommonExcep
//
//  File:         CommonExcep.cpp
//
//----------------------------------------------------------------------

// System Includes

// Local Includes

#include "../Common/CommonExcep.h"

// Macros

// Static initialization.

const CommonExcep CommonExcep::staticInitExcep;

//----------------------------------------------------------------------
// Method: Constructor
//----------------------------------------------------------------------
CommonExcep::CommonExcep() 
  :
     m_completeFlag (false)
{
   // Exercise all static accessors.

   defaultErrExcepInfo();
}

//----------------------------------------------------------------------
// Method: Constructor
//----------------------------------------------------------------------
CommonExcep::CommonExcep(const char* file,
                         const char* method,
                         const int   line,
                         const char* errMsg,
                         const int   errCode,
                         const char* subID) 
  :
     m_completeFlag (true)
{
   add(this->id(), file, method, line, errMsg, errCode, subID);
}

//----------------------------------------------------------------------
// Method: id
//----------------------------------------------------------------------
const char* CommonExcep::id() const     
{
   static const char* classId = "CommonExcep";
   return classId;
}

//----------------------------------------------------------------------
// Method: push
//----------------------------------------------------------------------
void CommonExcep::push(const char* file,
                       const char* method,
                       const int   line,
                       const char* errMsg,
                       const int   errCode)     
{
   add(this->id(), file, method, line, errMsg, errCode, "");
}

//----------------------------------------------------------------------
// Method: importPrior
//----------------------------------------------------------------------
void CommonExcep::importPrior(const CommonExcep& excep)    
{
   try
   {
      // Pull the exception info entries from the deque
      // of the prior exception and push them onto the
      // front of our deque.

      const ExcepInfoDeque& exceptions = excep.exceptions();

      ExcepInfoDequeRevIterConst iter;

      for (iter = exceptions.rbegin(); iter != exceptions.rend(); ++iter)
      {
         m_excepInfoDeque.push_front(*iter);
      }
   }

   catch(...)
   {
      m_completeFlag = false;
   }
}

//----------------------------------------------------------------------
// Method: Constructor
//----------------------------------------------------------------------
CommonExcep::CommonExcep(const char* id,
                         const char* file,
                         const char* method,
                         const int   line,
                         const char* errMsg,
                         const int   errCode,
                         const char* subID)     
  :
     m_completeFlag (true)
{
   add(id, file, method, line, errMsg, errCode, subID);
}

//----------------------------------------------------------------------
// Method: add
//----------------------------------------------------------------------
void CommonExcep::add(const char* id,
                      const char* file,
                      const char* method,
                      const int   line,
                      const char* errMsg,
                      const int   errCode,
                      const char* subID)     
{
   try
   {
      // If this is not the first exception in the deque, then
      // only push the exception information if it is coming from
      // a different file and method than the latest excep info.
      // Note: This prevents duplicate info from appearing in the
      // excep info stack from the same method. An exception should
      // only have one origination point within a given method.

      if (m_excepInfoDeque.size() > 0)
      {
         ExcepInfo& latestExcepInfo = m_excepInfoDeque.back();

         if (!strcmp(latestExcepInfo.file().c_str(),   file) && 
             !strcmp(latestExcepInfo.method().c_str(), method))
         {
            return;
         }
      }

      // Construct the exception info entry and push it on 
      // the back of the queue.

      ExcepInfo entry(id, file, method, line, errMsg, errCode, subID);
      m_excepInfoDeque.push_back(entry);
   }

   catch(...)
   {
      m_completeFlag = false;
   }
}

//----------------------------------------------------------------------
// Method: defaultErrExcepInfo
//----------------------------------------------------------------------
const const ExcepInfo& CommonExcep::defaultErrExcepInfo() const
{
   static const ExcepInfo  
        errExceptInfo
            (id(),
             __FILE__,
             "defaultErrExcepInfo",
             __LINE__,
             "Internal Excep Error: could not capture excep info.",
             0,
             "");

   return errExceptInfo;
}


