//----------------------------------------------------------------------
//
//  Copyright(c) 2001, Richard Nies
//
//  Permission granted to use this code without restriction as
//  long as this copyright notice appears in all source files.
//
//  Class:        CommonExcep
//
//  File:         CommonExcep.h
//
//  Description:  Defines a Common Exception object that can be used
//                to throw an exception condition with error information 
//                containing the fileid, method id and line where 
//                the exception was detected; along with an error 
//                message and error code if relevant.
// 
//                This object represents a stack of exceptions that 
//                can be used to trace an exception path up through a
//                chain of methods allowing tracing and logging 
//                responsibilities to be handled at higher levels 
//                without loss of information.
//
//                This class itself does not throw any exceptions 
//                (wouldn't want to have to force try/catch blocks
//                in every catch block!).
//
//----------------------------------------------------------------------
#ifndef _COMMONEXCEP_H
#define _COMMONEXCEP_H

// System Includes

// Local Includes

#include "../Common/ExcepInfo.h"
#include "../Common/ExcepInfoDeque.h"

// Macros

//----------------------------------------------------------------------
// Class: Common Exception (CommonExcep)
//----------------------------------------------------------------------
class CommonExcep
{
public:

   //---
   // Constructors/Destructors
   //---

   CommonExcep
      (const char* file,
       const char* method,
       const int   line,
       const char* errMsg   = "",
       const int   errCode  = 0,
       const char* subID    = "");

   virtual ~CommonExcep() {}

   //---
   // Accessor methods.
   //---
     
   const int             numEntries()  const;
   // Returns the number of entries on the exception deque.

   const ExcepInfo&      origExcep()   const;
   // Returns the original (first thrown) exception which resides
   // at the front of the deque.

   const ExcepInfo&      latestExcep() const;
   // Returns the latest exception encountered which resides at
   // the back of the deque.

   const ExcepInfoDeque& exceptions()  const;
   // Returns the deque of exception information indicating the
   // history of the exception as it has been caught and raised. 

   const bool            ok()          const;
   // Returns "true" if the exception is COMPLETE (did not encounter 
   // any internal errors in the process of having the excep info
   // added and being thrown). Returns "false" if the exception is
   // INCOMPLETE (an internal error was encountered (only exception 
   // loggers will probably be interested in this state).

   virtual const char*   id()          const;
   // Returns the unique text id to identify this exception.

   //---
   // Manipulator methods.
   //---
     
   void    push(const char* file,
                const char* method,
                const int   line,
                const char* errMsg = "");
   // Pushes the specified excep information onto the excep deque.

   void    push(const char* file,
                const char* method,
                const int   line,
                const char* errMsg,
                const int   errCode);
   // Pushes the specified excep information onto the excep deque.

   void    importPrior(const CommonExcep& excep);
   // Imports the specified excep information deque into this
   // exception as "prior exceptions" to the definition of this
   // exception.  This allows exceptions to be "filtered" into
   // other exceptions but import the prior exceptions history.

protected:

   //---
   // Constructors/Destructors
   //---

   CommonExcep(const char* id,
               const char* file,
               const char* method,
               const int   line,
               const char* errMsg,
               const int   errCode,
               const char* subID);
   // Used by derived exception classes.

   //---
   // Utility methods.
   //---
     
   void    add(const char* id,
               const char* file,
               const char* method,
               const int   line,
               const char* errMsg,
               const int   errCode,
               const char* subID);
   // Adds an excep info entry to the exceptions deque.

   //---
   // Data members.
   //---

   ExcepInfoDeque   m_excepInfoDeque;
   // Deque of exceptions where first (orig) exception thrown is at 
   // the front of the queue - latest thrown is at the back.  This is
   // used to conceptually implement the exception deque.

   bool             m_completeFlag;
   // Flag indicating if the exception is COMPLETE (no internal errors)
   // or INCOMPLETE (an internal exception error was encountered
   // - such as unable to push excep info onto the deque).

   //---
   // Static data member accessors.
   //---
     
   const ExcepInfo& defaultErrExcepInfo() const;
   // Returns the default exception info indicating that an
   // error occurred in capturing the exception info.

private:

   //---
   // Constructors/Destructors
   //---

   CommonExcep();
   // Used in the following private data member for static 
   // constant initialization.  This allows the static memory
   // accessors to be exercised on program startup instead of
   // at runtime.  This is desired for the exception class so that
   // memory allocation will be minimized during exception conditions.

   static const CommonExcep staticInitExcep;
   // Instance of CommonExcep used for exercising static accessors
   // in constructor to force program startup instantiation.
};  

//----------------------------------------------------------------------
// Inline Method: numEntries
//----------------------------------------------------------------------
inline
const int CommonExcep::numEntries()  const
{
   return m_excepInfoDeque.size();
}

//----------------------------------------------------------------------
// Inline Method: origExcep
//----------------------------------------------------------------------
inline
const ExcepInfo& CommonExcep::origExcep()  const
{
   if (m_excepInfoDeque.size() != 0)
      return m_excepInfoDeque.front();
   else
      return defaultErrExcepInfo();
}

//----------------------------------------------------------------------
// Inline Method: latestExcep
//----------------------------------------------------------------------
inline
const ExcepInfo& CommonExcep::latestExcep()  const
{
   if (m_excepInfoDeque.size() != 0)
      return m_excepInfoDeque.back();
   else
      return defaultErrExcepInfo();
}

//----------------------------------------------------------------------
// Inline Method: exceptions
//----------------------------------------------------------------------
inline
const ExcepInfoDeque& CommonExcep::exceptions() const
{
   return m_excepInfoDeque;
}

//----------------------------------------------------------------------
// Inline Method: ok
//----------------------------------------------------------------------
inline
const bool CommonExcep::ok() const
{
   return m_completeFlag;
}

//----------------------------------------------------------------------
// Method: push
//----------------------------------------------------------------------
inline
void CommonExcep::push(const char* file,
                       const char* method,
                       const int   line,
                       const char* errMsg)     
{
   push(file, method, line, errMsg, 0);
}

#endif



