/**************************************************************
Module : DataBreakpoint.cpp - set hardware breakpoint on memory
             Written 1998,1999 by Dmitri Leman
Purpose: NT kernel mode tracer.
**************************************************************/
extern "C"
{
#include "ntddk.h"
}
#include "HTrace.h"

#define DS_SELECTOR    0x23
#define FS_SELECTOR 0x30

bool    g_bISRInstalled = false;
int     g_iBreakpointIDTEntry = 0x1;
PVOID   g_pOldIDR;//Original interrupt handler

extern "C" ULONG KeSetAffinityThread
(
   PKTHREAD,
   ULONG affinityMask
);

void LogThreadSwap()
{
    static ULONG s_dwPrevThread[32];
    int l_iProcessor = KeGetCurrentProcessorNumber();
    static int s_iCount = 0;
    ULONG l_dwCurThreadID = (ULONG)PsGetCurrentThreadId();
    ULONG l_dwPrevThread = (ULONG)-1;
    if(l_iProcessor < 32)
    {
        l_dwPrevThread = s_dwPrevThread[l_iProcessor];
        s_dwPrevThread[l_iProcessor] = l_dwCurThreadID;
    }

    HTRACE(TG_ThreadSwaps, 
        "%d Swap %x->%x processor = %d. Hard brkpt",
        s_iCount++, l_dwPrevThread, l_dwCurThreadID, 
        l_iProcessor);
}//void LogThreadSwap

void _declspec(naked) InterruptHandler()
{
    _asm
    {  
        //Save all registers and set NT selectors
        pushad
        push    ds
        push    es
        push    gs
        push    fs
        mov     bx, DS_SELECTOR
        mov     ds, bx
        mov     es, bx
        mov     gs, bx
        mov     bx, FS_SELECTOR
        mov     fs, bx
        
        call    LogThreadSwap
        
        pop     fs
        pop     gs
        pop     es
        pop     ds
        popad

        //Pass control to the original ISR
        //jmp     cs:[g_pOldIDR]
        iretd
    }
}//void _declspec(naked) InterruptHandler

struct IdtEntry
{
    USHORT    m_wLowOffset;
    USHORT    m_wSelector;
    USHORT    m_wFlags;
        //P, DPL (2 bits), 0, D, 1, 1, 0, 00000000 -
        //P - present, D - 1 for 32 bit
    USHORT    m_wHighOffset;
};//struct IdtEntry

void HookInterrupt
(
    IdtEntry  * p_pIdtEntry,
    PVOID   p_pNewISR,
    PVOID * p_pOldISR
)
{
    //Disable interrupts on the current CPU
    KIRQL l_OldIRQL;
    KeRaiseIrql( HIGH_LEVEL, &l_OldIRQL);
    _asm    { 
        pushfd
        cli
    }

    if(p_pOldISR != NULL)
        *(PULONG)p_pOldISR = 
        (ULONG(p_pIdtEntry->m_wHighOffset) << 16) + 
            p_pIdtEntry->m_wLowOffset;

    //We don't touch flags or selector: only offset
    p_pIdtEntry->m_wLowOffset = USHORT(p_pNewISR);
    p_pIdtEntry->m_wHighOffset  = 
        USHORT(((ULONG)p_pNewISR) >> 16);

    //Enable interrupts back
    _asm    popfd
    KeLowerIrql(l_OldIRQL);
}//void HookInterrupt

void SetDataReadBreakpointAndISR
(
    bool    p_bSet,
    USHORT  p_wDataBreakPointSelector,
    ULONG   p_dwDataBreakPointOffset,
    PVOID   p_pNewISR,
    PVOID * p_pOldISR
)
{
    ULONG l_dwDR7;
    struct GdtAddr
    {
       USHORT m_wLimit;
       USHORT m_wBaseLow;
       USHORT m_wBaseHigh;
    };
    struct Descriptor
    {
       USHORT m_wLimit;
       USHORT m_wBaseLow;
       UCHAR  m_byBaseMid;
       UCHAR  m_byFlags1;
       UCHAR  m_byFlags2;
       UCHAR  m_byBaseHigh;
    };
    GdtAddr l_GdtAddr;
    GdtAddr l_IdtAddr;
    __asm { 
       mov eax, dr7
       mov l_dwDR7, eax
       lea eax, l_GdtAddr
       sgdt [eax]
       lea eax, l_IdtAddr
       sidt [eax]
    }
    ULONG l_dwBase = 0;
    Descriptor * l_pIDT = (Descriptor*)
        (ULONG(l_IdtAddr.m_wBaseHigh) << 16 | 
        l_IdtAddr.m_wBaseLow);

    if(p_bSet)
    {
        HookInterrupt(
            (IdtEntry*)l_pIDT +g_iBreakpointIDTEntry,
            p_pNewISR, p_pOldISR);

        l_dwDR7 |= (1 << 7);
            //Bit 7 - G3 - global break point enable
        l_dwDR7 |= (1 << 8) | (1 << 9);
            //LE GE for compatibility
        l_dwDR7 |= (1 << 28);//Set bit 28 and reset 
            //bit 29 to enable "write"
        l_dwDR7 &= ~(1 << 29);//breakpoint #3
        l_dwDR7 |= (1 << 30) || (1 << 31);//Len3 = 4

        Descriptor * l_pGDT = (Descriptor*)
            (ULONG(l_GdtAddr.m_wBaseHigh) << 16 | 
                l_GdtAddr.m_wBaseLow);
        Descriptor & l_rDescr = 
            l_pGDT[ p_wDataBreakPointSelector >> 3 ];
        l_dwBase   = ULONG(l_rDescr.m_byBaseHigh) << 24 | 
                     ULONG(l_rDescr.m_byBaseMid) << 16 |
                     ULONG(l_rDescr.m_wBaseLow);
        l_dwBase += p_dwDataBreakPointOffset;

        HTRACEK((KeyWordDriverDebug, 
            "Set hardware breakpoint at %x. Processor #%d",
            l_dwBase, KeGetCurrentProcessorNumber()));

    }
    else
    {
        l_dwDR7 &= ~(1 << 7);
            //Bit 7 - G3 - global break pointenable
        l_dwDR7 &= ~((1 << 8) | (1 << 9));
            //LE GE for compatibility
        l_dwDR7 &= ~(1 << 28);
            //Set bit 28 and reset bit 29 to enable "write"
        l_dwDR7 &= ~(1 << 29);
            //breakpoint #3
        l_dwDR7 &= ~((1 << 30) || (1 << 31));//Len3 = 4
    }

    __asm {
        mov eax, l_dwDR7
        mov DR7, eax
        mov eax, [l_dwBase]
        mov DR3, eax
    }
    if(!p_bSet)
    {
        HookInterrupt((IdtEntry*)l_pIDT + 
            g_iBreakpointIDTEntry, p_pNewISR, p_pOldISR);
    }
}//void SetDataReadBreakpointAndISR

bool SetBreakpointAndISR
(
    bool p_bSet
)
{
    if(g_bISRInstalled == p_bSet)
    {
        return FALSE;        
    }
   
    //We need to set breakpoints and install ISR for all
    //processors. I will use undocumented technique
    //presented by Jose Flores in February 2000 WDJ.
    //It uses undocumented KeSetAffinityThread to 
    //transfer the current thread to another processor.
    int l_iNumProcessors = *KeNumberProcessors;
    for(int l_iCPU = 0; l_iCPU < l_iNumProcessors; l_iCPU++)
    {
        PKTHREAD l_pThread = KeGetCurrentThread();
        KeSetAffinityThread(l_pThread, 1<<l_iCPU);

        //We set breakpoint on offset 0x124 within the 
        //current PCR (processor control block), which is 
        //pointed by fs selector. Offset 0x124 holds pointer 
        //to the current kernel thread.
        USHORT l_wFS = 0;
        ULONG l_dwOffset = 0x124;
        __asm 
        {
            mov ax, fs
            mov [l_wFS], ax
        }

        if (p_bSet)
        {
            SetDataReadBreakpointAndISR(true, l_wFS,l_dwOffset,
                    InterruptHandler, &g_pOldIDR);
        }
        else
        {
            SetDataReadBreakpointAndISR(false,l_wFS,l_dwOffset,
                    g_pOldIDR, NULL);
        }
    }//for(int l_iCPU = 0; l_iCPU < l_iNumProcessors; l_iCPU++)

    //Restore the affinity to run on all processors.
    PKTHREAD l_pThread = KeGetCurrentThread();
    KeSetAffinityThread(l_pThread, (1 << l_iNumProcessors)-1);

    g_bISRInstalled = p_bSet;
    return true;
}//bool SetBreakpointAn
