/**************************************************************
Module : DrvInit.cpp    - the main file of NT driver
             Written 1998,1999 by Dmitri Leman
Purpose: NT kernel mode tracer.
**************************************************************/
extern "C"
{
#include "ntddk.h"
}
#include "HTrace.h"

//Forward definitions:
NTSTATUS TracerDispatch
    ( IN PDEVICE_OBJECT p_pDeviceObject, IN PIRP p_pIrp );
VOID     TracerUnload( IN PDRIVER_OBJECT p_pDriverObject );


PDEVICE_OBJECT  g_pTracerDevice = NULL;//This device provides 
               //IOCTL interface with our launcher application

/*-------------------------------------------------------------

   FUNCTION: TracerDispatch

   PURPOSE:  Standard dispatch routine of the driver. Handles
    IRP_MJ_CREATE, IRP_MJ_CLOSE and IRP_MJ_DEVICE_CONTROL
      
   PARAMETERS:                   .
                                  
   IN PDEVICE_OBJECT   p_pDeviceObject - our device object
   IN PIRP             p_pIrp          - request packet

   RETURN VALUE:
      NTSTATUS 0 on success, error code otherwise
-------------------------------------------------------------*/

NTSTATUS TracerDispatch
( 
    IN PDEVICE_OBJECT   p_pDeviceObject,
    IN PIRP             p_pIrp 
)
{
    PIO_STACK_LOCATION  l_pIrpStack = 
            IoGetCurrentIrpStackLocation (p_pIrp);
    ULONG               l_ulIoControlCode = 
        l_pIrpStack->Parameters.DeviceIoControl.IoControlCode;

    //by default mark all requests as successful
    p_pIrp->IoStatus.Status      = STATUS_SUCCESS;
    p_pIrp->IoStatus.Information = 0;

    switch (l_pIrpStack->MajorFunction) {
    case IRP_MJ_CREATE:

        //HTRACEK(KeyWordDriverDebug, "Tracer: IRP_MJ_CREATE");

        break;

    case IRP_MJ_CLOSE:

        HTRACEK((KeyWordDriverDebug, "Tracer: IRP_MJ_CLOSE"));
        UnmapTraceMemory(0, l_pIrpStack->FileObject);
        break;

    case IRP_MJ_DEVICE_CONTROL:
        HTRACEK((KeyWordDriverDebug,
            "Tracer:IRP_MJ_DEVICE_CONTROL"));

        if(l_ulIoControlCode == IOCTL_TRACER_CHECK_ACTIVE ||
           l_ulIoControlCode == IOCTL_TRACER_DUMP_TRACE ||
           l_ulIoControlCode == IOCTL_TRACER_MAP_BUFFER ||
           l_ulIoControlCode == IOCTL_TRACER_UNMAP_BUFFER)
        {
            p_pIrp->IoStatus.Status = TraceIoctl(
                p_pDeviceObject, p_pIrp,
            IOCTL_TRACER_CHECK_ACTIVE, IOCTL_TRACER_DUMP_TRACE,
            IOCTL_TRACER_MAP_BUFFER,IOCTL_TRACER_UNMAP_BUFFER);
        }
        else
            p_pIrp->IoStatus.Status = STATUS_NOT_IMPLEMENTED;
        if(p_pIrp->IoStatus.Status == STATUS_NOT_IMPLEMENTED)
        {
            HTRACE(TG_Error, 
               "ERROR: Unknown IOCTL (%x)", l_ulIoControlCode);
        }
        break;
    }

    // Complete the IRP
    IoCompleteRequest( p_pIrp, IO_NO_INCREMENT );
    return p_pIrp->IoStatus.Status;
}//NTSTATUS TracerDispatch

extern bool SetBreakpointAndISR(bool p_bSet);

/*-------------------------------------------------------------

   FUNCTION: TracerUnload

   PURPOSE:  Called by the OS before unloading our driver
      
   PARAMETERS:                   .
                                  
   IN PDRIVER_OBJECT p_pDriverObject - our driver object

_____________________________________________________________*/

VOID TracerUnload
( 
    IN PDRIVER_OBJECT p_pDriverObject 
)
{
    HTRACEK((KeyWordDriverDebug, "Entering TracerUnload"));

    SetBreakpointAndISR(false);

    KeSetSwapContextNotifyRoutine(0);
    KeSetThreadSelectNotifyRoutine(0);

    WCHAR          l_szLinkName[]  = TRACER_DOS_DEVICE_NAME_W;
    UNICODE_STRING l_LinkName;

    // Delete the symbolic link for our device
    RtlInitUnicodeString( &l_LinkName, l_szLinkName );
    IoDeleteSymbolicLink( &l_LinkName );

    //Delete our device
    if ( g_pTracerDevice != p_pDriverObject->DeviceObject )
    {
        IoDeleteDevice( g_pTracerDevice ); 
        g_pTracerDevice = NULL;
    }
    IoDeleteDevice( p_pDriverObject->DeviceObject );

    HTRACEK((KeyWordDriverDebug, "Leaving TracerUnload"));

    //Uninitialize the trace
    EnableTraceToOutputLog(NULL, TG_Error, (ULONG)-1);
    TraceUnInitialize();
}//VOID TracerUnload


LOGICAL FASTCALL ThreadSelectNotifyRoutine
(
    IN HANDLE ThreadId
)
{
    
    static int s_iCount = 0;
    HTRACE(TG_ThreadSwaps, 
        "%d ThreadSelectNotifyRoutine(%x) proc = %d", 
        s_iCount++,
        ThreadId, KeGetCurrentProcessorNumber());
    
    return 0;
}

extern "C" NTSYSAPI NTSTATUS NTAPI 
ZwGetContextThread
(
    IN HANDLE ThreadHandle,
    OUT PCONTEXT Context
);

VOID FASTCALL SwapContextNotifyRoutine
(
    IN HANDLE OldThreadId,
    IN HANDLE NewThreadId
)
{
    
    static int s_iCount = 0;
    HTRACE(TG_ThreadSwaps, "%d Swap %x->%x processor = %d", 
            s_iCount++,
            OldThreadId, NewThreadId,
            KeGetCurrentProcessorNumber());
            
}//VOID FASTCALL SwapContextNotifyRoutine

/*-------------------------------------------------------------

   FUNCTION: NTSTATUS DriverEntry

   PURPOSE:  Standard entry point of NT kernel driver
      
   PARAMETERS:                   .
                                  
   IN PDRIVER_OBJECT  p_pDriverObject - our driver object
   IN PUNICODE_STRING p_pRegistryPath -our driver registry path

   RETURN VALUE:
      NTSTATUS      0 on success, error code otherwise
-------------------------------------------------------------*/
extern "C" NTSTATUS DriverEntry
(
    IN PDRIVER_OBJECT  p_pDriverObject, 
    IN PUNICODE_STRING p_pRegistryPath 
)
{
    //Initialize our trace immediately, so we will be able
    //to print error messages
    TraceInitialize();
    TraceAssignGroupsToStream
        (TO_DebugMonitor, TG_Error, (ULONG)-1);
    TraceAssignGroupKeyWordToStream(
        TO_DebugMonitor, KeyWordDriverDebug);
    EnableTraceToOutputLog
        (p_pDriverObject, TG_Error, (ULONG)-1);
    TraceAllocateBuffer(1024*1024, TG_ThreadSwaps, (ULONG)-1);


    KeSetSwapContextNotifyRoutine(SwapContextNotifyRoutine);
    KeSetThreadSelectNotifyRoutine(ThreadSelectNotifyRoutine);
    
    //Comment the following line to disable setting breakpoint
    //to thread switch.
    SetBreakpointAndISR(true);


    HTRACEK((KeyWordDriverDebug, "Entering DriverEntry"));

    NTSTATUS       l_Status;
    WCHAR          l_szDeviceName[]=TRACER_KERNEL_DEVICE_NAME;
    UNICODE_STRING l_DeviceName;
    WCHAR          l_szLinkName[]  = TRACER_DOS_DEVICE_NAME_W;
    UNICODE_STRING l_LinkName;

    //Initialize the device name
    RtlInitUnicodeString (&l_DeviceName, l_szDeviceName);

    //Create the device for communications with application
    l_Status = IoCreateDevice ( p_pDriverObject,
                                0,
                                &l_DeviceName,
                                FILE_DEVICE_TRACER,
                                0,
                                FALSE,//not exclusive
                                &g_pTracerDevice );

    if (!NT_SUCCESS(l_Status)) {
        HTRACE(TG_Error, "ERROR: IoCreateDevice(%*S)",
            l_DeviceName.Length, l_DeviceName.Buffer);
    }

    // If successful, make a symbolic link that allows for the 
    // device object's access from Win32 programs
    if (NT_SUCCESS(l_Status)) {

        // Create a symbolic link to be used by the application
        // as a file name to create a handle to our device
        RtlInitUnicodeString (&l_LinkName, l_szLinkName);
        l_Status = IoCreateSymbolicLink
            (&l_LinkName, &l_DeviceName);
        if (!NT_SUCCESS(l_Status)) 
        {
            HTRACE(TG_Error, "ERROR:IoCreateSymbolicLink(%*S)",
                l_LinkName.Length, l_LinkName.Buffer);
        }

        // Register dispatch routine
        p_pDriverObject->MajorFunction[IRP_MJ_CLOSE]      =
            p_pDriverObject->MajorFunction[IRP_MJ_CREATE] =
            p_pDriverObject->MajorFunction
                [IRP_MJ_DEVICE_CONTROL] = TracerDispatch;
        p_pDriverObject->DriverUnload = TracerUnload;
    }


    //In case of any error - uninitialize everything and return
    //the error to the OS, which will unload us
    if (!NT_SUCCESS(l_Status)) 
    {

        HTRACE(TG_Error, "ERROR: DriverEntry failed. %x ", 
            l_Status);

        RtlInitUnicodeString (&l_LinkName, l_szLinkName);
        IoDeleteSymbolicLink( &l_LinkName );

        if( g_pTracerDevice ) 
        {
            IoDeleteDevice( g_pTracerDevice );
            g_pTracerDevice = NULL;
        }
        TraceUnInitialize();
        return l_Status;
    }
    
    HTRACEK((KeyWordDriverDebug, "Leaving DriverEntry OK"));
    return l_Status;
}//NTSTATUS DriverEntry
