/**************************************************************
Module : InstDrv.cpp  - InstallDriver class to 
    register/unregister,start and stop drivers
         Written 1998,1999,2000 by Dmitri Leman
**************************************************************/

#include <windows.h>
#include "InstDrv.h"
#include "HTrace.h"

/*-------------------------------------------------------------

   FUNCTION: RegisterDriver

   PURPOSE:Registers a driver with the system configuration 
    manager
      
   PARAMETERS:                   .
                                  
    IN  LPCTSTR p_pszName       Name of the driver
    IN  LPCTSTR p_pszPath       Path of the driver file
   RETURN VALUE:
      bool      true on success, false on failure
-------------------------------------------------------------*/
bool InstallDriver::bRegisterDriver
(
    IN  LPCTSTR p_pszName,
    IN  LPCTSTR p_pszPath
)
{
    SC_HANDLE    l_hSCManager;

    l_hSCManager = OpenSCManager
        (NULL,NULL,SC_MANAGER_ALL_ACCESS);
    if(!l_hSCManager)
    {
        HTRACE(TG_Error,_T("ERROR: OpenSCManager failed - %s"),
            ERR_EXPL(GetLastError()));
        return false;
    }

    // Ignore success of installation: it may already be 
    // installed.
    bRegisterDriver(l_hSCManager, p_pszName, p_pszPath );

     CloseServiceHandle( l_hSCManager );
    return true;
}//bool InstallDriver::bRegisterDriver

/*-------------------------------------------------------------

   FUNCTION: InstallDriver::bRegisterDriver

   PURPOSE:Registers a driver with the system configuration 
   manager
      
   PARAMETERS:                   .
                                  
   IN SC_HANDLE p_hSchSCManager  service control manager handle
   IN LPCTSTR p_pszDriverName    Name of the driver 
   IN LPCTSTR p_pszDriverPath    Path to the driver executable

   RETURN VALUE:
      bool      true on success, false on failure
-------------------------------------------------------------*/
bool InstallDriver::bRegisterDriver
( 
    IN SC_HANDLE p_hSchSCManager, 
    IN LPCTSTR   p_pszDriverName, 
    IN LPCTSTR   p_pszDriverPath
)
{
    SC_HANDLE  l_hService = CreateService( 
        p_hSchSCManager, // SCManager database
        p_pszDriverName, // name of service
        p_pszDriverName, // name to display
        SERVICE_ALL_ACCESS,//desired access
        SERVICE_KERNEL_DRIVER,//service type
        SERVICE_DEMAND_START, // start type
        SERVICE_ERROR_NORMAL,  // error control type
        p_pszDriverPath,       // service's binary
        NULL,                  // no load ordering group
        NULL,                  // no tag identifier
        NULL,                  // no dependencies
        NULL,                  // LocalSystem account
        NULL                   // no password
        );
    if ( l_hService == NULL )
    {
        if(ERROR_SERVICE_EXISTS  == GetLastError())
        {
            return true;
        }

        HTRACE(TG_Error, 
            _T("ERROR: CreateService(%s,%s) failed - %s"), 
            p_pszDriverName, p_pszDriverPath, 
            ERR_EXPL(GetLastError()));
        return false;
    }
    CloseServiceHandle( l_hService );
    return true;
}//bool InstallDriver::bRegisterDriver

/*-------------------------------------------------------------

   FUNCTION: InstallDriver::bStartDriverAndOpenDevice

   PURPOSE:  Start the driver (should be already registered)
            and open a handle to the device
      
   PARAMETERS:                   .
                                  
   IN  LPCTSTR p_pszDriverName      Name of the driver 
   IN  LPCTSTR p_pszDeviceName      Name of the device
   OUT HANDLE *p_phDevice           Return handle to the device

   RETURN VALUE:
      bool      true on success, false on failure
-------------------------------------------------------------*/
bool InstallDriver::bStartDriverAndOpenDevice
(
   IN  LPCTSTR p_pszDriverName,
   IN  LPCTSTR p_pszDeviceName,
   OUT HANDLE *p_phDevice
)
{
    SC_HANDLE    l_hSCManager;
    bool        l_bResult;

    l_hSCManager = OpenSCManager
        (NULL,NULL,SC_MANAGER_ALL_ACCESS);
    if(!l_hSCManager)
    {
        HTRACE(TG_Error, 
            _T("ERROR: OpenSCManager failed - %s"), 
            ERR_EXPL(GetLastError()));
        return false;
    }

    // Ignore success of start: it may already be started.
    bStartDriver( l_hSCManager, p_pszDriverName);

    // Do make sure we can open it.
    l_bResult = bOpenDevice( p_pszDeviceName, p_phDevice );
    if(!l_bResult)
        bUnloadDriver(p_pszDriverName, true);

     CloseServiceHandle( l_hSCManager );

    return l_bResult;
}//bool InstallDriver::bStartDriverAndOpenDevice

/*-------------------------------------------------------------

   FUNCTION: InstallDriver::bStartDriver

   PURPOSE:  Start the driver(load to memory and start 
    execution)
      
   PARAMETERS:                   .
                                  
   IN SC_HANDLE p_hSchSCManager  service control manager handle
   IN  LPCTSTR p_pszDriverName   Name of the driver 

   RETURN VALUE:
      bool      true on success, false on failure
-------------------------------------------------------------*/
bool InstallDriver::bStartDriver
( 
    IN SC_HANDLE p_hSchSCManager,
    IN  LPCTSTR  p_pszDriverName
)
{
    SC_HANDLE  l_hService;
    bool       l_bResult;

    l_hService = OpenService( p_hSchSCManager,
                              p_pszDriverName,
                              SERVICE_ALL_ACCESS
                              );
    if ( l_hService == NULL )
    {
        HTRACE(TG_Error,_T("ERROR: OpenService(%s) failed-%s"),
            p_pszDriverName, ERR_EXPL(GetLastError()));
        return false;
    }

    l_bResult = StartService( l_hService, 0, NULL )
       || GetLastError() == ERROR_SERVICE_ALREADY_RUNNING;

    if(!l_bResult)
    {
        HTRACE(TG_Error,_T("ERROR:StartService(%s) failed-%s"),
            p_pszDriverName, ERR_EXPL(GetLastError()));
    }

    CloseServiceHandle( l_hService );

    return l_bResult;
}//bool InstallDriver::bStartDriver

/*-------------------------------------------------------------

   FUNCTION: InstallDriver::bOpenDevice

   PURPOSE:  Open the device and (optionally) return the handle
      
   PARAMETERS:                   .
                                  
   IN  LPCTSTR  p_pszDeviceName   - device name
   OUT HANDLE * p_phDevice        - return handle     

   RETURN VALUE:
      bool      true on success, false on failure
-------------------------------------------------------------*/
bool InstallDriver::bOpenDevice
( 
    IN LPCTSTR p_pszDeviceName, 
    OUT HANDLE *p_phDevice 
)
{
    HANDLE   l_hDevice = CreateFile( p_pszDeviceName,
                          GENERIC_READ | GENERIC_WRITE,
                          FILE_SHARE_READ | FILE_SHARE_WRITE,
                          NULL,
                          OPEN_EXISTING,
                          FILE_ATTRIBUTE_NORMAL,
                          NULL
                          );
    if ( l_hDevice == ((HANDLE)-1) )
    {
        HTRACE(TG_Error,_T("ERROR: CreateFile(%s) failed -%s"),
            p_pszDeviceName, ERR_EXPL(GetLastError()));
        return false;
    }

    if (p_phDevice )
        *p_phDevice  = l_hDevice;
    else
        CloseHandle(l_hDevice);
    return true;
}//bool InstallDriver::bOpenDevice

/*-------------------------------------------------------------

   FUNCTION: InstallDriver::bUnloadDriver

   PURPOSE:  Stop the driver (unload it from memory) and
             (optionally) unregister it.
      
   PARAMETERS:                   .
                                  
    IN LPCTSTR p_pszDriverName   Name of the driver 
    IN bool    p_bUnregister     Unregister or not

   RETURN VALUE:
      bool      true on success, false on failure
-------------------------------------------------------------*/
bool InstallDriver::bUnloadDriver
( 
    IN LPCTSTR p_pszDriverName, 
    IN bool    p_bUnregister 
)
{
    SC_HANDLE    l_hSCManager = OpenSCManager(NULL, NULL, 
        SC_MANAGER_ALL_ACCESS);
    if(!l_hSCManager)
    {
        HTRACE(TG_Error, 
            _T("ERROR: OpenSCManager failed - %s"), 
            ERR_EXPL(GetLastError()));
        return false;
    }

    bStopDriver(l_hSCManager, p_pszDriverName);
    if(p_bUnregister)
    {
        bRemoveDriver(l_hSCManager, p_pszDriverName);
    }
    CloseServiceHandle(l_hSCManager);
    return true;
}//bool InstallDriver::bUnloadDriver

/*-------------------------------------------------------------

   FUNCTION: InstallDriver::bStopDriver

   PURPOSE:  Stop the driver (unload from memory)
      
   PARAMETERS:                   .
                                  
   IN SC_HANDLE p_hSchSCManager  service control manager handle
   IN LPCTSTR   p_pszDriverName  Name of the driver 

   RETURN VALUE:
      bool      true on success, false on failure
-------------------------------------------------------------*/
bool InstallDriver::bStopDriver
( 
    IN SC_HANDLE p_hSchSCManager,
    IN LPCTSTR   p_pszDriverName
)
{
    SC_HANDLE       l_hService;
    bool            l_bResult;
    SERVICE_STATUS  serviceStatus;

    l_hService = OpenService( p_hSchSCManager, p_pszDriverName,
        SERVICE_ALL_ACCESS );
    if ( l_hService == NULL )
    {
        HTRACE(TG_Error,_T("ERROR: OpenService(%s) failed-%s"),
            p_pszDriverName, ERR_EXPL(GetLastError()));
        return false;
    }


    l_bResult = ControlService(l_hService,SERVICE_CONTROL_STOP,
        &serviceStatus )? true : false;
    if(!l_bResult)
    {
      HTRACE(TG_Error,_T("ERROR:ControlService(%s) failed-%s"),
            p_pszDriverName, ERR_EXPL(GetLastError()));
    }
    CloseServiceHandle( l_hService );
    return l_bResult;
}//bool InstallDriver::bStopDriver

/*-------------------------------------------------------------

   FUNCTION: InstallDriver::RemoveDriver

   PURPOSE:  Unregister the driver
      
   PARAMETERS:                   .
                                  
   IN SC_HANDLE p_hSchSCManager  control manager handle
   IN LPCTSTR   p_pszDriverName          

   RETURN VALUE:
      bool      true on success, false on failure
-------------------------------------------------------------*/
bool InstallDriver::bRemoveDriver
( 
    IN SC_HANDLE p_hSchSCManager, 
    IN LPCTSTR   p_pszDriverName 
)
{
    SC_HANDLE  l_hService;
    bool       l_bResult;

    l_hService = OpenService( p_hSchSCManager, p_pszDriverName,
        SERVICE_ALL_ACCESS);
    if(!l_hService)
    {
        HTRACE(TG_Error,_T("ERROR: OpenService(%s) failed-%s"),
            p_pszDriverName, ERR_EXPL(GetLastError()));
        return false;
    }
    l_bResult = DeleteService( l_hService )? true : false;
    if(!l_bResult)
    {
       HTRACE(TG_Error,_T("ERROR:DeleteService(%s) failed-%s"),
            p_pszDriverName, ERR_EXPL(GetLastError()));
    }
    CloseServiceHandle( l_hService );
    return l_bResult;
}//bool InstallDriver::bRemoveDriver
