/**************************************************************
Module : TrcView.cpp      - trace buffer viewer
             Written 2000 by Dmitri Leman
Purpose: A part of the tracer framework.
**************************************************************/
#if defined WIN32
#include <windows.h>
#include <stdio.h>
#include "resource.h"
#include "HTrace.h"

#else if defined(__linux__)

#include <unistd.h>
#include <curses.h>
#include <malloc.h>
#include <string.h>
#include <sys/time.h>
#include "HTrace.h"

//This module was ported from WIN32 to ncurses.
//Therefore, I included few declarations, which allows
//some WIN32-specific to compile under linux.
typedef WINDOW * HDC;
typedef WINDOW * HWND;
typedef int BOOL;
#define __int64 long long
typedef struct {int left, top, right, bottom;} RECT;
void DrawText(
    WINDOW * p_pWindow, const char * p_pszBuffer,
    int p_iNumBytes, RECT * p_pClientRect, int p_iFlags)
{
    int l_iRes = 
        mvwaddnstr(stdscr, 0,0, p_pszBuffer, p_iNumBytes);
    wclrtobot(p_pWindow);
}
static ULONG GetTickCount()
{
    clock_t l_Clock = clock();
    return (ULONG)(((__int64)l_Clock) * 1000 / CLOCKS_PER_SEC);
}
#define SB_VERT 1
#define SB_THUMBTRACK 1
#define SB_LINEUP     2
#define SB_LINEDOWN   3
#define SB_PAGEUP     4
#define SB_PAGEDOWN   5
#define SB_TOP        6
#define SB_BOTTOM     7

#define _snprintf snprintf
#endif


static const char s_szCaption[] = "Trace Viewer";

static char s_szFullINIPath[_MAX_PATH];

#define MAX_ROW_LEN     256
#define NUM_EXTRA_ROWS  2

struct BufferViewer
{
    LocalTraceBufferPointers * m_pBufferPointers;

    HWND    m_hWnd;

#ifdef WIN32
    HFONT   m_hFont;
    int     m_iFontHeight;
    int     m_iFontWidth;
    int     m_iLogicalWidthChars;
    int     m_iLogicalWidthPixels;
#endif

    char *  m_pBuffer;
    int     m_iBufferSize;
    int     m_iNumLines;
    int     m_iFirstPaintIndex;
    int     m_iNumBytesPainted;
    int     m_iFirstViewPosWhenPainted;
    int     m_iNumLinesPainted;

    int     m_iByteShownFirst;
    bool    m_bStickToEnd;
    bool    m_bMoveToStartNextTime;

    RECT    m_ClientRect;
    int     m_iNumLinesInClientArea;

    BufferViewer(HWND p_hWnd)
    {
        m_hWnd = p_hWnd;
        m_pBufferPointers = NULL;
        #ifdef WIN32
        m_hFont = 0;
        m_iFontHeight = 0;
        m_iFontWidth = 0;
        m_iLogicalWidthChars = 0;
        m_iLogicalWidthPixels = 0;
        #endif
        m_pBuffer = NULL;
        m_iBufferSize = 0;
        m_iNumLines = 0;
        m_iFirstPaintIndex = -1;
        m_iNumBytesPainted = -1;
        m_iFirstViewPosWhenPainted = -1;
        m_iNumLinesPainted = 0;
        m_iByteShownFirst = 0;
        m_bStickToEnd = true;
        m_bMoveToStartNextTime = false;
        m_iNumLinesInClientArea = 0;
        #ifdef __linux__
        m_iInvalidated = 0;
        m_iScrollMin = 0;
        m_iScrollMax = 0;
        m_iScrollPos = 0; 
        m_pWndStatus = NULL;
        m_pWndHelp = NULL;             
        #endif
    }
    void PaintBufferWithCopy
    (
        HDC  p_hDC,
        const char * p_pTraceBuffer,
        int p_lBufferSize,
        int p_iFirstViewPos,
        int p_iCurByte,
        int p_iSpaceAfterScrollPos, 
        int p_iSpaceBeforeScrollPos
    );
    void PaintBuffer
    (
        int                 p_iFirstViewPos,
        HDC                 p_hDC,
        RECT              * p_pClientRect
    );
    void OnPaint();
    void OnSizeChanges();
    void OnTimerMessage();
    void OnVScrollMessage
    (
        int p_nScrollCode,
        int p_nPos,
        int p_nTrackPos
    );
    #ifdef __linux__
    //Few more routines to allow portions of WIN32 specific
    //code to run under ncurses
    void InvalidateRect(HWND,RECT*,BOOL)
    {
        m_iInvalidated++;
    }
    void GetScrollRange(HWND, int, int * p_piMin,int * p_piMax)
    {
        *p_piMin = m_iScrollMin;
        *p_piMax = m_iScrollMax;
    } 
    void SetScrollRange(HWND, int,int p_iMin, int p_iMax,BOOL)
    {
        m_iScrollMin = p_iMin;
        m_iScrollMax = p_iMax;    
    }
    void SetScrollPos(HWND, int, int p_iPos,BOOL)
    {
        m_iScrollPos = p_iPos;
    }       
    int m_iInvalidated;
    int m_iScrollMin;
    int m_iScrollMax;
    int m_iScrollPos;
    WINDOW * m_pWndStatus;
    WINDOW * m_pWndHelp;
    #endif
};//struct BufferViewer

/*-------------------------------------------------------------

   FUNCTION: BufferViewer::PaintBufferWithCopy

   PURPOSE:  Worker routine to copy a portion of a trace buffer
    to a local display buffer and paint it to the screen.
   PARAMETERS:
    HDC  p_hDC                  - device context
    const char * p_pTraceBuffer - the whole circular buffer
    int p_lBufferSize           - total size of the buffer
    int p_iFirstViewPos         - current scroll position
    int p_iCurByte        - index of the byte, which should be
                                  painted on top of the window.
    int p_iSpaceAfterScrollPos  - number of valid bytes in the 
                            buffer before p_iCurByte
    int p_iSpaceBeforeScrollPos - number of valid bytes in the 
                            buffer after p_iCurByte
-------------------------------------------------------------*/
void BufferViewer::PaintBufferWithCopy
(
    HDC  p_hDC,
    const char * p_pTraceBuffer,
    int p_lBufferSize,
    int p_iFirstViewPos,
    int p_iCurByte,
    int p_iSpaceAfterScrollPos, 
    int p_iSpaceBeforeScrollPos
)
{
    int l_iSize = (m_iNumLinesInClientArea + NUM_EXTRA_ROWS) * 
        MAX_ROW_LEN;
    if(m_iNumLines < m_iNumLinesInClientArea || !m_pBuffer)
    {
        //The worst uncertainty is the line width.
        //We will assume MAX_ROW_LEN = 256 byte line width.
        //Hopefully, most of lines will be shorter.
        //If most of lines will be longer than 256,
        //we will print fewer lines, than fit in window
        char * p_Buff = (char * )malloc(l_iSize);
        if(p_Buff == NULL)
        {
            HTRACE(TG_Error, 
                "ERROR: malloc(%d) failed", l_iSize);
            return;
        }
        free(m_pBuffer);
        m_pBuffer = p_Buff;
        m_iBufferSize = l_iSize;
        m_iNumLines = m_iNumLinesInClientArea;
    }//if(m_iNumLines < m_iNumLinesInClientArea)

    //Now we need to copy memory from the trace buffer to the
    //screen buffer.
    //We need to start before the current position to find the 
    //beginning of the line, which contains the current 
    //position.
    int l_iLookBack = MAX_ROW_LEN;
    if(l_iLookBack + p_iSpaceAfterScrollPos < l_iSize)
    {  
        //If we have not enough data after the scroll point,
        //we need to look farther behind to avoid leaving blank
        //space
        l_iLookBack = l_iSize - p_iSpaceAfterScrollPos;
    }
    if(l_iLookBack > p_iSpaceBeforeScrollPos)
       l_iLookBack = p_iSpaceBeforeScrollPos;
    int l_iStartFromByteWithExtra = p_iCurByte - l_iLookBack;
    if(l_iStartFromByteWithExtra < 0)
       l_iStartFromByteWithExtra += p_lBufferSize;

    int l_iDisplaySize = p_iSpaceAfterScrollPos + l_iLookBack;
    if(l_iDisplaySize > l_iSize)
       l_iDisplaySize = l_iSize;
    int l_iLookForward = l_iDisplaySize;
    if(l_iStartFromByteWithExtra + l_iLookForward > 
        p_lBufferSize)
    {
        l_iLookForward = p_lBufferSize - 
            l_iStartFromByteWithExtra;
        memcpy(m_pBuffer, p_pTraceBuffer + 
            l_iStartFromByteWithExtra, l_iLookForward);
        //The display area continues at the beginning of buffer
        memcpy(m_pBuffer + l_iLookForward, p_pTraceBuffer, 
            l_iDisplaySize - l_iLookForward);
    }
    else
    {
        memcpy(m_pBuffer, p_pTraceBuffer + 
            l_iStartFromByteWithExtra, l_iLookForward);
    }

    int i;
    int l_iNumLines = 0;
    int l_iBeginning = l_iLookBack, l_iEnd = 0;

    //First count lines after the current byte, which is at the
    //position l_iLookBack in the m_pBuffer
    l_iEnd = l_iDisplaySize;//if we will not find newlines
    for(i = l_iLookBack; i < l_iDisplaySize; i++)
    {
        if(m_pBuffer[i] == '\n')
        {
            l_iEnd = i;
            l_iNumLines++;
            if(l_iNumLines >= m_iNumLinesInClientArea)
                break;
        }
    }
    //Next look back until we find the beginning of the line,
    //which the current position belongs to. And look further
    //back if we have not enough lines already to fill screen.
    for(i = l_iLookBack-1; i >= 0; i--)
    {
        if(m_pBuffer[i] == '\n')
        {
            l_iBeginning = i+1;
            l_iNumLines++;
            if(l_iNumLines >= m_iNumLinesInClientArea)
                break;
        }
    }
    if(i < 0)
        l_iBeginning = 0;

    m_iFirstPaintIndex = l_iBeginning;
    m_iNumBytesPainted = l_iEnd - l_iBeginning;
    DrawText(p_hDC, m_pBuffer + m_iFirstPaintIndex, 
                    m_iNumBytesPainted, &m_ClientRect, 0);

    m_iFirstViewPosWhenPainted = p_iFirstViewPos;
    m_iNumLinesPainted = l_iNumLines;
}//void BufferViewer::PaintBufferWithCopy

/*------------------------------------------------------------

   FUNCTION: BufferViewer::PaintBuffer

   PURPOSE:  Performs actual painting of the trace buffer
   PARAMETERS:
    p_iFirstViewPos - current byte (between 0 and the buffer 
        size) to be shown at the top of the window (measured 
        from the initial start of writing)
------------------------------------------------------------*/
void BufferViewer::PaintBuffer
(
    int                 p_iFirstViewPos,
    HDC                 p_hDC,
    RECT              * p_pClientRect
)
{
    if(!m_pBufferPointers)
        return;
    ULONG l_dwBytesWritten = 
       m_pBufferPointers->m_pGlobalFooter->m_dwNumBytesWritten;
    ULONG l_dwBufferSize = m_pBufferPointers->m_dwTextAreaSize;

    if(p_iFirstViewPos > (int)l_dwBytesWritten)
        p_iFirstViewPos = l_dwBytesWritten;
        //don't show empty space

    char * l_pText = m_pBufferPointers->m_pTextArea;

    //Decide which byte in the trace buffer corresponds to the
    //scroll position. This will be the last line to be shown.

    int l_iSpaceAfterScrollPos = 
        l_dwBytesWritten - p_iFirstViewPos;
    if(l_iSpaceAfterScrollPos > (int)l_dwBufferSize)
    {
        //The scroll position is behind the beginning of the 
        //buffer.
        if(m_iFirstViewPosWhenPainted == p_iFirstViewPos &&
           m_iFirstPaintIndex >= 0 && m_pBuffer != NULL)
        {
            //Fortunately, we already have the data in the 
            //display buffer
            DrawText(p_hDC, m_pBuffer + m_iFirstPaintIndex, 
                        m_iNumBytesPainted, p_pClientRect, 0);
            return;
        }

        DrawText(p_hDC, "Data Lost", 9, p_pClientRect, 0);
        return;
    }
    int l_iSpaceBeforeScrollPos;
    if(l_dwBytesWritten > l_dwBufferSize)
       l_iSpaceBeforeScrollPos = 
        l_dwBufferSize - l_iSpaceAfterScrollPos;
    else
       l_iSpaceBeforeScrollPos = p_iFirstViewPos;

    int l_iStartFromByte = p_iFirstViewPos % l_dwBufferSize;

    PaintBufferWithCopy(
        p_hDC, l_pText, l_dwBufferSize, 
        p_iFirstViewPos, l_iStartFromByte, 
        l_iSpaceAfterScrollPos, l_iSpaceBeforeScrollPos);
}//void BufferViewer::PaintBuffer

#ifdef WIN32 //{
/*-------------------------------------------------------------

   FUNCTION: BufferViewer::OnPaint

   PURPOSE:  Called to paint screen buffer view window
-------------------------------------------------------------*/
void BufferViewer::OnPaint()
{
    PAINTSTRUCT l_Paint;
    BeginPaint(m_hWnd, &l_Paint);

    if(m_pBufferPointers)
    {
        HGDIOBJ l_hPrevFont = SelectObject
            (l_Paint.hdc, m_hFont);
        if(!m_iFontHeight)
        {
            SIZE l_Size;
            m_iFontHeight = GetTextExtentPoint32(
                l_Paint.hdc, "W", 4, &l_Size);
            m_iFontHeight = l_Size.cy;
            m_iFontWidth  = l_Size.cx;
            m_iLogicalWidthPixels = m_iLogicalWidthChars * 
                m_iFontWidth;

            OnSizeChanges();
        }
        //long l_lVertPos = GetScrollPos(m_hWnd, SB_VERT);
        FillRect(l_Paint.hdc, &m_ClientRect, 
            (HBRUSH) GetStockObject(WHITE_BRUSH));
        
        if(m_bMoveToStartNextTime && m_iByteShownFirst < 
           (int) (m_pBufferPointers->m_pGlobalFooter->
           m_dwNumBytesWritten - 
           m_pBufferPointers->m_dwTextAreaSize))
        {
            m_iByteShownFirst = 
                m_pBufferPointers->m_pGlobalFooter->
                    m_dwNumBytesWritten - 
                m_pBufferPointers->m_dwTextAreaSize;
        }
        m_bMoveToStartNextTime = false;
        PaintBuffer(m_iByteShownFirst, 
            l_Paint.hdc, &m_ClientRect);
        SelectObject(l_Paint.hdc, l_hPrevFont);
    }

    EndPaint(m_hWnd, &l_Paint);
}//void BufferViewer::OnPaint()

/*-------------------------------------------------------------

   FUNCTION: BufferViewer::OnSizeChanges

   PURPOSE:  Called in response to WM_SIZE
-------------------------------------------------------------*/
void BufferViewer::OnSizeChanges()
{
    GetClientRect(m_hWnd, &m_ClientRect);
    m_iNumLinesInClientArea = 0;
    if(m_iFontHeight > 0)
    {
        m_iNumLinesInClientArea = 
        (m_ClientRect.bottom - m_ClientRect.top)/m_iFontHeight;
    }
    InvalidateRect(m_hWnd, NULL, FALSE);
}//void BufferViewer::OnSizeChanges()
#endif//WIN32 }

/*-------------------------------------------------------------

   FUNCTION: BufferViewer::OnTimerMessage

   PURPOSE:  Called in response to WM_TIMER message in the 
    buffer view window. Calculates a new scroll position and 
    forces repaint
------------------------------------------------------------*/
void BufferViewer::OnTimerMessage()
{
    if(!m_pBufferPointers)
        return;

    int l_iBytesWritten = m_pBufferPointers->
        m_pGlobalFooter->m_dwNumBytesWritten;
    int l_iBufferSize = m_pBufferPointers->m_dwTextAreaSize;
    int l_iRange = 0;
    int l_iScrollPos = 0;
    //We will set the scroll range depending where the 
    //current viewing point (m_iByteShownFirst) is. If it
    //is within the memory buffer, we need to set the range
    //to the size of the buffer (or the number of bytes 
    //written, if the buffer is not full yet). If the 
    //current viewing point is behind the buffer (the data 
    //were lost), we need to set the range to the whole 
    //distance between the newest data and the view point.

    if(l_iBytesWritten < l_iBufferSize)
    {//The buffer is not full yet.
        l_iRange = l_iBytesWritten;
        l_iScrollPos = m_iByteShownFirst;
    }
    else
    {//The buffer is full
        int l_iFirstByteInBuffer = 
            l_iBytesWritten - l_iBufferSize;
        if(m_iByteShownFirst < l_iFirstByteInBuffer)
        {   //The view is behind the oldest data in the 
            //buffer. Set the range to the whole distance 
            //between the newest data and the view point.
            l_iRange = l_iBytesWritten - m_iByteShownFirst;
            l_iScrollPos = 0;
        }
        else
        {
            //The view point is within the buffer.
            //The range should be equal to the buffer size.
            l_iRange = l_iBufferSize;
            l_iScrollPos = m_iByteShownFirst - 
                l_iFirstByteInBuffer;
        }
    }
    if(m_bStickToEnd)
    {
        l_iScrollPos = l_iRange;
        m_iByteShownFirst = l_iBytesWritten;
    }
    if(m_iByteShownFirst > l_iBytesWritten)
    {
       m_iByteShownFirst = 0;//buffer was reset
       m_bStickToEnd = true;
    }

    SetScrollRange(m_hWnd, SB_VERT, 0, l_iRange, FALSE);
    SetScrollPos(m_hWnd, SB_VERT, l_iScrollPos, TRUE);

    if(m_iByteShownFirst == m_iFirstViewPosWhenPainted &&
       m_iNumLinesPainted >= m_iNumLinesInClientArea)
    {
       return;//don't repaint the same place again
    }

    InvalidateRect(m_hWnd, NULL, FALSE);
}//void BufferViewer::OnTimerMessage

void BufferViewer::OnVScrollMessage
(
    int p_nScrollCode,
    int p_nPos,
    int p_nTrackPos
)
{
    int l_iMin = 0, l_iMax = 0;
    GetScrollRange(m_hWnd, SB_VERT, &l_iMin, &l_iMax);
    long l_lCurPos = p_nPos;
    long l_lNewPos = p_nPos;
    
    switch(p_nScrollCode)
    {
    case SB_THUMBTRACK:
        l_lNewPos = p_nTrackPos;
        if(p_nTrackPos >= l_iMax)
        {
            m_bStickToEnd = true;
        }
        break;
    case SB_LINEUP:
        l_lNewPos -= 80;
        break;
    case SB_LINEDOWN:
        l_lNewPos += 80;
        break;
    case SB_PAGEUP:
        l_lNewPos -= 22*80;
        break;
    case SB_PAGEDOWN:
        l_lNewPos += 22*80;
        break;
    case SB_TOP:
        l_lNewPos = 0;
        m_bMoveToStartNextTime = true;
        break;
    case SB_BOTTOM:
        l_lNewPos = l_iMax;
        m_bStickToEnd = true;
        break;
    }
    if(l_lNewPos <  l_iMin)
       l_lNewPos =  l_iMin;
    else if(l_lNewPos >  l_iMax)
       l_lNewPos = l_iMax;
       
    int l_iDiffPos = l_lNewPos - l_lCurPos;
    m_iByteShownFirst += l_iDiffPos;

    if(l_iDiffPos < 0)
    {
        m_bStickToEnd = false;
    }

    if(m_iByteShownFirst < (int)
        (m_pBufferPointers->m_pGlobalFooter->m_dwNumBytesWritten-
        m_pBufferPointers->m_dwTextAreaSize))
    {
        m_bMoveToStartNextTime = true;
    }
    else if(m_iByteShownFirst >= (int)
        m_pBufferPointers->m_pGlobalFooter->m_dwNumBytesWritten)
    {
        m_bStickToEnd = true;
    }
    SetScrollPos(m_hWnd, SB_VERT, l_lNewPos, TRUE);
}//void BufferViewer::OnVScrollMessage

#ifdef WIN32 //{
/*-------------------------------------------------------------

   FUNCTION: WindowProc

   PURPOSE:  Window procedure for the trace buffer view window
------------------------------------------------------------*/
LRESULT CALLBACK WindowProc
(  
    HWND p_hWnd,
    UINT p_uMsg,
    WPARAM p_wParam,
    LPARAM p_lParam
)
{
    HTRACEK((KeyWordAppDebug, "View WindowProc %x %x %x %x", 
        p_hWnd,  p_uMsg,  p_wParam, p_lParam ));

    long l_lBufferViewer = GetWindowLong(p_hWnd, 0);
    BufferViewer * l_pViewer =(BufferViewer * )l_lBufferViewer;
    switch(p_uMsg)
    {
    case WM_CREATE:
        {
            SetTimer(p_hWnd, 1, 1000, NULL);

            BufferViewer * l_pViewer =new BufferViewer(p_hWnd);
            l_pViewer->m_pBufferPointers = pGetTraceBuffer();
            l_pViewer->m_hFont = 
                (HFONT)GetStockObject(SYSTEM_FIXED_FONT);

            SetWindowLong(p_hWnd, 0, (long)l_pViewer);

            if(l_pViewer->m_pBufferPointers)
            {
                SetScrollRange(p_hWnd, SB_VERT, 0, 
                 l_pViewer->m_pBufferPointers->m_dwTextAreaSize,
                    FALSE);
                l_pViewer->m_iLogicalWidthChars = 512;
                SetScrollRange(p_hWnd, SB_HORZ, 0, 
                    l_pViewer->m_iLogicalWidthChars, FALSE);
            }
        }
    case WM_SIZE:
        if(l_pViewer)
            l_pViewer->OnSizeChanges();
        break;
    case WM_TIMER:
        if(l_pViewer)
            l_pViewer->OnTimerMessage();
        break;
    case WM_VSCROLL:
        if(l_pViewer)
        {
            int l_nScrollCode = (int) LOWORD(p_wParam);
            short int l_nPos = (short int) HIWORD(p_wParam);
            HWND l_hwndScrollBar = (HWND) p_lParam;

            SCROLLINFO l_ScrollInfo;
            memset(&l_ScrollInfo, 0, sizeof(l_ScrollInfo));
            l_ScrollInfo.cbSize = sizeof(l_ScrollInfo);
            l_ScrollInfo.fMask = 
                SIF_TRACKPOS | SIF_POS | SIF_RANGE;
            GetScrollInfo(p_hWnd, SB_VERT, &l_ScrollInfo);
            long l_lPos = l_ScrollInfo.nPos;
            l_pViewer->OnVScrollMessage(l_nScrollCode, 
                l_ScrollInfo.nPos, l_ScrollInfo.nTrackPos);
        }
        break;
    case WM_PAINT:
        if(l_pViewer)
            l_pViewer->OnPaint();
        break;
    case WM_USER:
        KillTimer(p_hWnd, 1);
        SetTimer(p_hWnd, 1, p_wParam, NULL);
        break;
    case WM_USER+1:
        if(l_pViewer && l_pViewer->m_pBufferPointers && 
           l_pViewer->m_pBufferPointers->m_pGlobalFooter)
           l_pViewer->m_pBufferPointers->m_pGlobalFooter->
                m_dwNumBytesWritten = 0;
        break;
    case WM_DESTROY:
        KillTimer(p_hWnd, 1);
        PostQuitMessage(0);
        
        if(l_pViewer)
        {
            l_pViewer->m_pBufferPointers = NULL;
            delete l_pViewer;
        }
        SetWindowLong(p_hWnd, 0, 0);
        break;
    }
    return DefWindowProc(p_hWnd, p_uMsg, p_wParam, p_lParam);
}//LRESULT CALLBACK WindowProc

/*-------------------------------------------------------------

   FUNCTION: OnStreamComboBoxCommand

   PURPOSE:  Called when the User selects a trace output stream
-------------------------------------------------------------*/
void OnStreamComboBoxCommand(HWND p_hWndDialog)
{
    int l_iSelectedOutput = 
        SendDlgItemMessage(p_hWndDialog, IDC_COMBO_OUTPUT, 
            CB_GETITEMDATA, SendDlgItemMessage(p_hWndDialog, 
            IDC_COMBO_OUTPUT, CB_GETCURSEL,0,0), 0);
    if(l_iSelectedOutput <= 0)
        return;
    ULONG  l_ulGroups = 
        TraceGetAssignedGroupsToStream(l_iSelectedOutput);
    LPCSTR l_pszKeyWords = 
     TraceGetAssignedGroupKeyWordsToStream(l_iSelectedOutput);

    #define SET_CHECK(ControlID, GroupID) \
      SendDlgItemMessage(p_hWndDialog, ControlID, BM_SETCHECK,\
      (l_ulGroups & GroupID)? BST_CHECKED : BST_UNCHECKED, 0);
    SET_CHECK(IDC_CHECK_BANK_ACCOUNT, TG_BankAccount)
    SET_CHECK(IDC_CHECK_THREAD_SWAPS, TG_ThreadSwaps)
    SET_CHECK(IDC_CHECK_MESSAGE_BOX , TG_MessageBox )
    SET_CHECK(IDC_CHECK_ERRORS      , TG_Error      )
    char l_szMask[20];
    sprintf(l_szMask, "0x%x", l_ulGroups);
    
    SetDlgItemText(p_hWndDialog,IDC_EDIT_GROUP_MASK, l_szMask);
    SetDlgItemText(p_hWndDialog,IDC_EDIT_GROUP_KEYWORDS, 
        l_pszKeyWords);

    SetDlgItemText(p_hWndDialog, IDC_WARNING, "");
}//void OnStreamComboBoxCommand()

/*-------------------------------------------------------------

   FUNCTION: ShowWarning

   PURPOSE:  Show a help line in the Warning box with more
    information about the selected trace output.
-------------------------------------------------------------*/
void ShowWarning(int p_iSelectedOutput, HWND p_hWndDialog)
{
    if(p_iSelectedOutput == TO_MemoryBuffer)
    {
        SetDlgItemText(p_hWndDialog, IDC_WARNING, 
            "New mask is shared by all applications");
    }
    else if(p_iSelectedOutput > 0)
    {
        SetDlgItemText(p_hWndDialog, IDC_WARNING, 
          "To apply new mask to other apps and drivers "
          "you need to restart them (only memory buffer mask "
          "is shared and effective immediately");
    }
    else
    {
        SetDlgItemText(p_hWndDialog, IDC_WARNING, 
            "Please, select a trace output stream "
            "in the combo box");
    }

}

/*------------------------------------------------------------

   FUNCTION: OnGroupChanged

   PURPOSE:Called when the user clicks on a check box to enable
    or disable a single trace output stream
-------------------------------------------------------------*/
void OnGroupChanged
    (HWND p_hWndDialog, int p_iControlID, int p_iGroupID)
{
    int l_iSelectedOutput = 
        SendDlgItemMessage(p_hWndDialog, IDC_COMBO_OUTPUT, 
            CB_GETITEMDATA, SendDlgItemMessage(p_hWndDialog, 
            IDC_COMBO_OUTPUT, CB_GETCURSEL,0,0), 0);
    if(l_iSelectedOutput > 0)
    {
        int l_iChecked = SendDlgItemMessage
            (p_hWndDialog, p_iControlID, BM_GETCHECK,0,0);
        //The following call will set or reset only 
        //the single group ID bit == p_iGroupID
        TraceAssignGroupsToStream(l_iSelectedOutput, 
        (l_iChecked == BST_CHECKED)? p_iGroupID:0, p_iGroupID);
        OnStreamComboBoxCommand(p_hWndDialog);//Update view
    }
    ShowWarning(l_iSelectedOutput, p_hWndDialog);
}//void OnGroupChanged

/*-------------------------------------------------------------

   FUNCTION: OnSetButton

   PURPOSE:  Apply new settings entered by the User in the 
   dialog, such as group key mask, group keywords, trace file 
   name and view refresh rate
-------------------------------------------------------------*/
void OnSetButton(HWND p_hWndDialog)
{
    SetDlgItemText(p_hWndDialog, IDC_WARNING, "");

    int l_iSelectedOutput = 
        SendDlgItemMessage(p_hWndDialog, IDC_COMBO_OUTPUT, 
            CB_GETITEMDATA, SendDlgItemMessage(p_hWndDialog, 
            IDC_COMBO_OUTPUT, CB_GETCURSEL,0,0), 0);
    char l_szText[512];
    if(l_iSelectedOutput > 0)
    {
        if(l_iSelectedOutput == TO_File)
        {
            GetDlgItemText(p_hWndDialog, IDC_EDIT_FILE_TRACE, 
                l_szText, sizeof(l_szText));
            TraceSetOutputToFile(l_szText, 0, 0, 0,0);
        }

        GetDlgItemText(p_hWndDialog, IDC_EDIT_GROUP_MASK, 
            l_szText, sizeof(l_szText));
        ULONG l_ulNewMask = strtoul(l_szText, NULL, 0);
        TraceAssignGroupsToStream(l_iSelectedOutput, 
            l_ulNewMask, (DWORD)-1);

        GetDlgItemText(p_hWndDialog, IDC_EDIT_GROUP_KEYWORDS, 
            l_szText, sizeof(l_szText));
        TraceAssignGroupKeyWordToStream
            (l_iSelectedOutput, l_szText);

        OnStreamComboBoxCommand(p_hWndDialog);//Update view

        TraceReadWriteSettingsFromIniFile(s_szFullINIPath,true,
            l_iSelectedOutput);
    }
    ShowWarning(l_iSelectedOutput, p_hWndDialog);

    GetDlgItemText(p_hWndDialog, IDC_EDIT_RATE, l_szText, 
        sizeof(l_szText));
    long l_lNewRate = strtol(l_szText, NULL, 0);
    if(l_lNewRate > 0)
        SendDlgItemMessage(p_hWndDialog, IDC_VIEWER, WM_USER, 
        l_lNewRate, 0);

}//void OnSetButton(HWND p_hWndDialog)

/*-------------------------------------------------------------

   FUNCTION: OnDumpButton

   PURPOSE:  Dump content of trace buffer to disk in response 
             to click on Dump button
-------------------------------------------------------------*/
void OnDumpButton(HWND p_hWndDialog)
{
    char l_szText[_MAX_PATH];
    GetDlgItemText(p_hWndDialog, IDC_EDIT_FILE_DUMP, l_szText,
        sizeof(l_szText));
    char l_szPath[_MAX_PATH];
    GetFullPathName(l_szText,sizeof(l_szPath), l_szPath, NULL);
    if(GetFileAttributes(l_szPath) != -1)
    {
        char l_szMessage[512];
        sprintf(l_szMessage, "File %s already exists. "
            "Do you want to overwrite?", l_szPath);
            
        if(IDNO == MessageBox(p_hWndDialog, l_szMessage, 
            s_szCaption, MB_YESNO))
        {
            SetDlgItemText(p_hWndDialog, IDC_WARNING, 
                "File dumping aborted");
            return;
        }
    }//if(GetFileAttributes(l_szPath) != -1)
    WritePrivateProfileString("Trace", "TraceDumpFile",
        l_szPath, s_szFullINIPath);

    if(!TraceDumpBufferToFile(l_szPath))
    {
        SetDlgItemText(p_hWndDialog, IDC_WARNING, 
            "File dumping failed");
    }
}//void OnDumpButton()
#endif //WIN32 }

//struct RecordOverhead is used to configure a single overhead
//measurement test of a single output stream
struct RecordOverhead
{
    const char * m_pszName;
    DWORD        m_dwStreamID;
    int          m_iNumLoops;
    bool         m_bAddNewLineToNoFormat;    
    double       m_dFormatOverheadAverageMs;
    long         m_lFormatOverheadAverageCycles;
    long         m_lFormatOverheadMinCycles;
    double       m_dNoFormatOverheadAverageMs;
    long         m_lNoFormatOverheadAverageCycles;
    long         m_lNoFormatOverheadMinCycles;
};

#ifdef WIN32
//RDTSC instruction loads the current value of 
//the processor's timestamp counter into the EDX:EAX
//OpCode 0x0f31
#define GetProcessorCyclesCount(p_i64ReturnCounter) \
    __asm push eax                                  \
    __asm push edx                                  \
    __asm RDTSC                                     \
    __asm mov DWORD PTR p_i64ReturnCounter, EAX     \
    __asm mov DWORD PTR p_i64ReturnCounter + 4, EDX \
    __asm pop eax                                   \
    __asm pop edx                                         
#else
#define GetProcessorCyclesCount(p_i64ReturnCounter) \
      __asm__ __volatile__ (                        \
      "push %%eax;"                                 \
      "push %%edx;"                                 \
      "RDTSC;"                                      \
      "movl %%EAX, %0;"                             \
      "movl %%EDX, 4+%0;"                           \
      "pop %%eax;"                                  \
      "pop %%edx": "=m"(p_i64ReturnCounter): );
#endif

/*-------------------------------------------------------------

   FUNCTION: MeasureOverheadOfSingleStream

   PURPOSE:Run loops and measure overhead of a single 
    output stream
      
-------------------------------------------------------------*/
void MeasureOverheadOfSingleStream
(
    RecordOverhead & p_Record
)
{
    if(p_Record.m_iNumLoops <= 0 || p_Record.m_pszName == NULL)
        return;
    int i;
    //Make sure that TG_MeasOverhead is not enabled for
    //any stream
    for(i = 0; i < 32; i++)
    {
        //Clear TG_MeasOverhead bit for l_dwStream
        DWORD l_dwStream = 1 << i;
       TraceAssignGroupsToStream(l_dwStream,0,TG_MeasOverhead);
    }
    //Now enable TG_MeasOverhead for p_dwStreamID
    TraceAssignGroupsToStream(p_Record.m_dwStreamID, 
        TG_MeasOverhead, TG_MeasOverhead);

    //We will do two 3 measurements-1 dummy(without the trace),
    //the second - with HTRACE macro with 1 parameter
    //and the third - with OutputTraceString function.
    unsigned __int64 l_ui64TotalCyclesInDummy = 0;
    ULONG            l_ulMinCyclesInDummy = (ULONG)-1;
    ULONG            l_ulMaxCyclesInDummy = 0;
    long l_lTickStartDummy = 0, l_lTickEndDummy = 0;

    unsigned __int64 l_ui64TotalCyclesInTraceFormat = 0;
    ULONG            l_ulMinCyclesInTraceFormat = (ULONG)-1;
    ULONG            l_ulMaxCyclesInTraceFormat = 0;
    long l_lTickStartTraceFormat = 0,l_lTickEndTraceFormat = 0;

    unsigned __int64 l_ui64TotalCyclesInTraceNoFormat = 0;
    ULONG            l_ulMinCyclesInTraceNoFormat = (ULONG)-1;
    ULONG            l_ulMaxCyclesInTraceNoFormat = 0;
    long l_lTickStartTraceNoFormat=0,l_lTickEndTraceNoFormat=0;


    int l_iNumLoops = p_Record.m_iNumLoops;
    //This is the dummy loop
    l_lTickStartDummy = GetTickCount();
    for(i = 0; i < l_iNumLoops; i++)
    {
        unsigned __int64 l_ui64CyclesStart;
        GetProcessorCyclesCount(l_ui64CyclesStart);
        //Dummy operation
        unsigned __int64 l_ui64CyclesEnd;
        GetProcessorCyclesCount(l_ui64CyclesEnd);
        
        ULONG l_ulCyclesInside =
            (ULONG)(l_ui64CyclesEnd - l_ui64CyclesStart);
        l_ui64TotalCyclesInDummy += l_ulCyclesInside;
        if(l_ulMinCyclesInDummy > l_ulCyclesInside)
           l_ulMinCyclesInDummy = l_ulCyclesInside;
        if(l_ulMaxCyclesInDummy < l_ulCyclesInside)
           l_ulMaxCyclesInDummy = l_ulCyclesInside;
    }
    l_lTickEndDummy = GetTickCount();

    //This is the trace loop using HTRACE
    l_lTickStartTraceFormat = GetTickCount();
    for(i = 0; i < l_iNumLoops; i++)
    {
        unsigned __int64 l_ui64CyclesStart;
        GetProcessorCyclesCount(l_ui64CyclesStart);
        HTRACE(TG_MeasOverhead, "%3.3d Trace\n", (i&0xff));
        unsigned __int64 l_ui64CyclesEnd;
        GetProcessorCyclesCount(l_ui64CyclesEnd);
        
        ULONG l_ulCyclesInside =
            (ULONG)(l_ui64CyclesEnd - l_ui64CyclesStart);
        l_ui64TotalCyclesInTraceFormat += l_ulCyclesInside;
        if(l_ulMinCyclesInTraceFormat > l_ulCyclesInside)
           l_ulMinCyclesInTraceFormat = l_ulCyclesInside;
        if(l_ulMaxCyclesInTraceFormat < l_ulCyclesInside)
           l_ulMaxCyclesInTraceFormat = l_ulCyclesInside;
    }
    l_lTickEndTraceFormat = GetTickCount();

    //This is the trace loop using OutputTraceString
    char l_szText[15] = "*** Trace\n";
    if(!p_Record.m_bAddNewLineToNoFormat)
        l_szText[9] = ' ';
    l_lTickStartTraceNoFormat = GetTickCount();
    for(i = 0; i < l_iNumLoops; i++)
    {
        unsigned __int64 l_ui64CyclesStart;
        GetProcessorCyclesCount(l_ui64CyclesStart);
        l_szText[0] = (i%8) + '0';
        OutputTraceString(TG_MeasOverhead, l_szText, 10);
        unsigned __int64 l_ui64CyclesEnd;
        GetProcessorCyclesCount(l_ui64CyclesEnd);
        
        ULONG l_ulCyclesInside =
            (ULONG)(l_ui64CyclesEnd - l_ui64CyclesStart);
        l_ui64TotalCyclesInTraceNoFormat += l_ulCyclesInside;
        if(l_ulMinCyclesInTraceNoFormat > l_ulCyclesInside)
           l_ulMinCyclesInTraceNoFormat = l_ulCyclesInside;
        if(l_ulMaxCyclesInTraceNoFormat < l_ulCyclesInside)
           l_ulMaxCyclesInTraceNoFormat = l_ulCyclesInside;
    }
    l_lTickEndTraceNoFormat = GetTickCount();

    //Disable TG_MeasOverhead for p_dwStreamID
    TraceAssignGroupsToStream(
        p_Record.m_dwStreamID, TG_MeasOverhead, 0);

    long l_lAverCyclesDummy = 
        (long)(l_ui64TotalCyclesInDummy / l_iNumLoops);
    long l_lAverCyclesTraceFormat = 
        (long)(l_ui64TotalCyclesInTraceFormat / l_iNumLoops);
    long l_lAverCyclesTraceNoFormat = 
        (long)(l_ui64TotalCyclesInTraceNoFormat / l_iNumLoops);
    long l_lTotalTimeDummy = l_lTickEndDummy-l_lTickStartDummy;
    long l_lTotalTimeTraceFormat = 
            l_lTickEndTraceFormat - l_lTickStartTraceFormat;
    long l_lTotalTimeTraceNoFormat = 
            l_lTickEndTraceNoFormat -l_lTickStartTraceNoFormat;

    HTRACEK(("AppD", "%s Dummy Loops %d Total ms %d "
        "MinCycles %d AverCycles %d\n",
        p_Record.m_pszName, l_iNumLoops, l_lTotalTimeDummy,
        l_ulMinCyclesInDummy, l_lAverCyclesDummy));

    HTRACEK(("AppD", 
        "%s Format Trace Loops %d Total ms %d "
        "MinCycles %d AverCycles %d\n",
        p_Record.m_pszName, l_iNumLoops, 
        l_lTotalTimeTraceFormat,
        l_ulMinCyclesInTraceFormat, l_lAverCyclesTraceFormat));

    HTRACEK(("AppD", 
        "%s No Format Trace Loops %d Total ms %d "
        "MinCycles %d AverCycles %d\n",
        p_Record.m_pszName, l_iNumLoops, 
        l_lTotalTimeTraceNoFormat,
        l_ulMinCyclesInTraceNoFormat, 
        l_lAverCyclesTraceNoFormat));

    p_Record.m_dFormatOverheadAverageMs = 
        (double(l_lTotalTimeTraceFormat - l_lTotalTimeDummy))/
        l_iNumLoops;
    p_Record.m_lFormatOverheadAverageCycles = 
        l_lAverCyclesTraceFormat - l_lAverCyclesDummy;
    p_Record.m_lFormatOverheadMinCycles = 
        l_ulMinCyclesInTraceFormat - l_ulMinCyclesInDummy;
    
    p_Record.m_dNoFormatOverheadAverageMs = 
        (double(l_lTotalTimeTraceNoFormat-l_lTotalTimeDummy))/
        l_iNumLoops;
    p_Record.m_lNoFormatOverheadAverageCycles = 
        l_lAverCyclesTraceNoFormat - l_lAverCyclesDummy;
    p_Record.m_lNoFormatOverheadMinCycles = 
        l_ulMinCyclesInTraceNoFormat - l_ulMinCyclesInDummy;

    HTRACEK(("AppD", 
        "%s Format Trace Overhead ms %d MinTicks %d "
        "AverTicks %d\n",
        p_Record.m_pszName, 
        l_lTotalTimeTraceFormat - l_lTotalTimeDummy,
        p_Record.m_lFormatOverheadAverageCycles,
        p_Record.m_lFormatOverheadMinCycles));
    HTRACEK(("AppD", 
        "%s NoForm Trace Overhead ms %d MinTicks %d "
        "AverTicks %d\n",
        p_Record.m_pszName, 
        l_lTotalTimeTraceNoFormat - l_lTotalTimeDummy,
        p_Record.m_lNoFormatOverheadAverageCycles,
        p_Record.m_lNoFormatOverheadMinCycles));
    return;
}//void MeasureOverheadOfSingleStream

/*-------------------------------------------------------------

   FUNCTION: MeasureOverhead

   PURPOSE:  Run loops to measure overhead of various 
    trace methods
      
-------------------------------------------------------------*/
char * MeasureOverhead(char * p_pszText)
{
    RecordOverhead l_Records[] = {
    {"TO_DebugMonitor", TO_DebugMonitor, 1000, true},
    {"TO_MemoryBuffer", TO_MemoryBuffer, 100 * 1000, true},
    {"TO_Console"     , TO_Console     , 10 * 1000, false},
    {"TO_File"        , TO_File        , 10 * 1000, true},
    {"TO_File with flush",TO_File|TO_FlushFile,1 * 1000,true},
    };
    TraceSetOutputToFile(p_pszText, 0, 0, 0,0);

    char * l_pszBuffer = (char*)malloc(1024);
    int l_iAllocated = 1024;
    int l_iFilled = 0;
    int i;
    for(i = 0; i < sizeof(l_Records)/sizeof(l_Records[0]); i++)
    {
        RecordOverhead & l_rRecord = l_Records[i];
        MeasureOverheadOfSingleStream(l_rRecord);
    }
    
    TraceSetOutputToFile("OverheadResult.txt", TO_DebugMonitor,
        (ULONG)-1, 0,0);
    
    strcpy(l_pszBuffer, "\n\n\nThese are results of calling\n"
        "HTRACE(\"%3.3d Trace\\n\", counter); and\n"
        "OutputTraceString(\"*** Trace\\n\");\n");
    l_iFilled = strlen(l_pszBuffer);
    
    OutputTraceStringUnconditional(TO_DebugMonitor | TO_File, 
        l_pszBuffer, l_iFilled);     
    
    for(i = 0; i < sizeof(l_Records)/sizeof(l_Records[0]); i++)
    {
        RecordOverhead & l_rRecord = l_Records[i];
   
        char l_szAddMessage1[100];
        char l_szAddMessage2[100];
        _snprintf(l_szAddMessage1, sizeof(l_szAddMessage1),
        "%s HTRACE overhead ms %lf AverTicks %d MinTicks %d\n",
            l_rRecord.m_pszName,
            l_rRecord.m_dFormatOverheadAverageMs,
            l_rRecord.m_lFormatOverheadAverageCycles,
            l_rRecord.m_lFormatOverheadMinCycles);
        _snprintf(l_szAddMessage2, sizeof(l_szAddMessage2),
            "%s OutputTraceString overhead ms %lf "
            "AverTicks %d MinTicks %d\n",
            l_rRecord.m_pszName,
            l_rRecord.m_dNoFormatOverheadAverageMs,
            l_rRecord.m_lNoFormatOverheadAverageCycles,
            l_rRecord.m_lNoFormatOverheadMinCycles);
        int l_iLen1 = strlen(l_szAddMessage1);
        int l_iLen2 = strlen(l_szAddMessage2);
        if(l_iFilled + l_iLen1 + l_iLen2 >= l_iAllocated)
        {
            int l_iSize = l_iAllocated+l_iLen1+l_iLen2+1024;
            char * p = (char*)realloc(l_pszBuffer, l_iSize);
            if(p == NULL)
                break;
            l_pszBuffer = p;
            l_iAllocated = l_iSize;
        }
        memcpy(l_pszBuffer + l_iFilled, l_szAddMessage1, 
            l_iLen1+1);
        l_iFilled += l_iLen1; 
        memcpy(l_pszBuffer + l_iFilled, l_szAddMessage2, 
            l_iLen2+1);
        l_iFilled += l_iLen2;
        OutputTraceStringUnconditional(
            TO_DebugMonitor | TO_File, 
            l_szAddMessage1, strlen(l_szAddMessage1)); 
        OutputTraceStringUnconditional(
            TO_DebugMonitor | TO_File, 
            l_szAddMessage2, strlen(l_szAddMessage2));         
    }//for(i = 0; i < sizeof(l_Records)/sizeof(l_Records[0])
   return l_pszBuffer;
}//char * MeasureOverhead

#ifdef WIN32 //{
/*------------------------------------------------------------

   FUNCTION: OnDialogCommand

   PURPOSE:  Process WM_COMMAND messages in the dialog box
      
------------------------------------------------------------*/
void OnDialogCommand
(
    HWND p_hWndDialog,
    WORD p_wNotifyCode,
    WORD p_wID,
    HWND p_hWndCtl
)
{
    static int l_iInside = 0;
    l_iInside++;
    switch(p_wID)
    {
    case IDOK:
        EndDialog(p_hWndDialog, 1);
        break;
    case IDCANCEL:
        EndDialog(p_hWndDialog, 0);
        break;
    case IDC_COMBO_OUTPUT:
        OnStreamComboBoxCommand(p_hWndDialog);
        break;
    case IDC_CHECK_BANK_ACCOUNT:
        OnGroupChanged(p_hWndDialog, IDC_CHECK_BANK_ACCOUNT, 
            TG_BankAccount);
        break;
    case IDC_CHECK_THREAD_SWAPS:
        OnGroupChanged(p_hWndDialog, IDC_CHECK_THREAD_SWAPS, 
            TG_ThreadSwaps);
        break;
    case IDC_CHECK_ERRORS:
        OnGroupChanged(p_hWndDialog, IDC_CHECK_ERRORS      , 
            TG_Error);
        break;
    case IDC_CHECK_MESSAGE_BOX:
        OnGroupChanged(p_hWndDialog, IDC_CHECK_MESSAGE_BOX , 
            TG_MessageBox);
        break;
    case IDC_EDIT_GROUP_MASK:
    case IDC_EDIT_GROUP_KEYWORDS:
    case IDC_EDIT_RATE:
        if(p_wNotifyCode == EN_CHANGE && l_iInside == 1)
        {
            SetDlgItemText(p_hWndDialog, IDC_WARNING, 
                "Click \"Set\" to set the new value");
        }
        else if(p_wNotifyCode == EN_SETFOCUS &&
            p_wID == IDC_EDIT_GROUP_KEYWORDS)
        {
            SetDlgItemText(p_hWndDialog, IDC_WARNING, 
                "This demo supports 2 keywords: \"AppD\" and"
                "\"DrvD\" ");
        }

        break;
    case IDC_BUTTON_SET:
        OnSetButton(p_hWndDialog);
        break;
    case IDC_BUTTON_DUMP:
        OnDumpButton(p_hWndDialog);
        break;
    case IDC_BUTTON_OVERHEAD:
        if(IDYES == MessageBox(p_hWndDialog, 
            "Do you want to start test loop to "
            "measure trace overhead?", s_szCaption, MB_YESNO))
        {
            char l_szText[_MAX_PATH];
            GetDlgItemText(p_hWndDialog, IDC_EDIT_FILE_TRACE, 
                l_szText, sizeof(l_szText));
        
            MeasureOverhead(l_szText);
        }
        break;
    case IDC_BUTTON_CLEAR:
        SendDlgItemMessage(p_hWndDialog, IDC_VIEWER, WM_USER+1,
            0,0);
        break;
    }
    l_iInside--;
}//void OnDialogCommand

/*-------------------------------------------------------------

   FUNCTION: DialogProc

   PURPOSE:  Dialog box message procedure.
      
-------------------------------------------------------------*/
int CALLBACK DialogProc
(  
    HWND   p_hWnd,
    UINT   p_uMsg,
    WPARAM p_wParam,
    LPARAM p_lParam
)
{
    static ls_DlgProcCount = 0;
    HTRACEK((KeyWordAppDebug, "%d DialogProc %x %x %x %x", 
     ls_DlgProcCount++, p_hWnd,  p_uMsg,  p_wParam, p_lParam));
    switch(p_uMsg)
    {
    case WM_INITDIALOG:
    {
        #define ADD_STREAM(StreamID)                     \
            SendDlgItemMessage(p_hWnd, IDC_COMBO_OUTPUT, \
                CB_SETITEMDATA,                          \
                SendDlgItemMessage(p_hWnd, IDC_COMBO_OUTPUT, \
                CB_ADDSTRING, 0, (LPARAM)#StreamID),         \
                StreamID);

        ADD_STREAM(TO_DebugMonitor);
        ADD_STREAM(TO_File);
        ADD_STREAM(TO_FlushFile);
        ADD_STREAM(TO_MessageBox);
        ADD_STREAM(TO_Console);
        ADD_STREAM(TO_MemoryBuffer);
        ADD_STREAM(TO_OutputLog);

        char l_szPath[_MAX_PATH];
        GetPrivateProfileString("Trace", "TraceDumpFile",
            "TraceDump.txt", 
            l_szPath, sizeof(l_szPath), s_szFullINIPath);
        SetDlgItemText(p_hWnd, IDC_EDIT_FILE_DUMP, l_szPath);

        LPCSTR l_pszFileName = TraceGetCurTraceFileName();
        if(l_pszFileName == NULL || *l_pszFileName == 0)
           l_pszFileName = "Trace.txt";
        SetDlgItemText(p_hWnd, IDC_EDIT_FILE_TRACE, 
            l_pszFileName);
        return TRUE;
    }
    case WM_COMMAND:
    {
        WORD l_wNotifyCode = HIWORD(p_wParam);
        WORD l_wID = LOWORD(p_wParam);
        HWND l_hwndCtl = (HWND) p_lParam;

        OnDialogCommand(p_hWnd, l_wNotifyCode,l_wID,l_hwndCtl);
    }
    break;
    case WM_SIZE:
    {
        RECT l_rDlgClient, l_rViewer;
        GetClientRect(p_hWnd, &l_rDlgClient);
        HWND l_hWndView = GetDlgItem(p_hWnd, IDC_VIEWER);
        GetWindowRect(l_hWndView, &l_rViewer);
        MapWindowPoints(NULL, p_hWnd, (POINT*)(&l_rViewer), 2);
        int l_iWidth  = l_rDlgClient.right - 2*l_rViewer.left;
        int l_iHeight = l_rDlgClient.bottom - 
            l_rViewer.top - l_rViewer.left;
        //Resize the viewer leaving top unchanged and adjusting
        //right and bottom to make equal margin.
        if(l_iWidth > 100 && l_iHeight > 100)
        {
            SetWindowPos(l_hWndView, NULL, 
                l_rViewer.left, l_rViewer.top, 
                l_iWidth, l_iHeight, 
                SWP_NOZORDER | SWP_NOMOVE | SWP_NOACTIVATE);
        }
    }
    }
    return FALSE;
}//LRESULT CALLBACK DialogProc

/*-------------------------------------------------------------

   FUNCTION: ShowDialog

   PURPOSE:  Show modal dialog box with the trace buffer view
      
   RETURN VALUE:
      BOOL TRUE on success, FALSE on failure
-------------------------------------------------------------*/
BOOL ShowDialog()
{
    WNDCLASS l_Class;
    memset(&l_Class,0,sizeof(l_Class));
    l_Class.lpfnWndProc = WindowProc;
    l_Class.hCursor = LoadCursor(0,IDC_ARROW);
    l_Class.hbrBackground = (HBRUSH)(1 + COLOR_WINDOW);
    l_Class.lpszClassName = "TraceViewer";
    l_Class.cbWndExtra = 4;
    RegisterClass(&l_Class);
    
    DialogBox(NULL, MAKEINTRESOURCE(IDD_DIALOG_VIEWER), NULL, 
        DialogProc);
    return TRUE;
}//BOOL ShowDialog()
#endif
#if defined(__linux__)
WINDOW * CreateWindow
(int p_iHeight, int p_iWidth, int p_iY, int p_iX)
{
    WINDOW * l_pWnd = newwin(p_iHeight, p_iWidth, p_iY, p_iX);
    if(l_pWnd == NULL)
    {
        HTRACE(TG_Error,"ERROR:newwin(%d, %d, %d, %d) failed", 
            p_iHeight, p_iWidth, p_iY, p_iX);
        return NULL;
    }
    return l_pWnd;
}

BOOL CreateWindows
(BufferViewer * p_pViewer, int p_iWidth, int p_iHeight)
{
    if(p_iHeight < 3)
    {
        HTRACE(TG_Error, "ERROR: too few rows in the window");
        return FALSE;
    }
    int l_iMainWndHeight = p_iHeight-2;
    p_pViewer->m_hWnd = 
        CreateWindow(l_iMainWndHeight, p_iWidth, 0,0);
    p_pViewer->m_pWndStatus = 
        CreateWindow(1, p_iWidth, l_iMainWndHeight,0);   
    p_pViewer->m_pWndHelp = 
        CreateWindow(1, p_iWidth, l_iMainWndHeight+1,0);     
    if(p_pViewer->m_hWnd == NULL || 
        p_pViewer->m_pWndStatus == NULL || 
       p_pViewer->m_pWndHelp == NULL)
    {
        HTRACE(TG_Error, "ERROR: failed to create window");
        return FALSE;
    }
    p_pViewer->m_iNumLinesInClientArea = l_iMainWndHeight;
    return TRUE;
}//BOOL CreateWindows

void DeleteWindows(BufferViewer * p_Viewer)
{
    delwin(p_Viewer->m_hWnd);
    p_Viewer->m_hWnd = NULL;
    delwin(p_Viewer->m_pWndStatus);
    p_Viewer->m_pWndStatus = NULL;
    delwin(p_Viewer->m_pWndHelp);
    p_Viewer->m_pWndHelp = NULL;
}

BOOL ShowDialog()
{
    initscr();
    cbreak();
    noecho();
    keypad(stdscr, TRUE);
    
    int l_iWindowWidth = COLS;
    int l_iWindowHeight = LINES;
    int l_iRefreshDelay = 1000;
    BufferViewer * l_pViewer = new BufferViewer(NULL);
    
    if(!CreateWindows(l_pViewer, 
        l_iWindowWidth, l_iWindowHeight))
    {
        endwin();
        delete l_pViewer;
        return FALSE;
    }
    refresh();
    l_pViewer->m_pBufferPointers = pGetTraceBuffer();
        
    char l_szHelp[] = "Use arrows to scroll, q - stop, "
        "+/- - change rate, o - measure overhead";
    
    mvwprintw(l_pViewer->m_pWndHelp, 0,0, l_szHelp); 
    wrefresh(l_pViewer->m_pWndHelp);
         
    bool l_bContinue = true;
    long l_lTimeRepaint = 0;
    wtimeout(stdscr, l_iRefreshDelay);    
    while(l_bContinue)
    {
        int l_iChar = wgetch(stdscr);
        if(l_iChar != ERR)
        {
            switch(l_iChar)
            {
                case 's': case 'q':
                    l_bContinue = false;
                    break;//happy end
                case KEY_UP:
                    l_pViewer->OnVScrollMessage(SB_LINEUP, 
                        l_pViewer->m_iScrollPos,
                        l_pViewer->m_iScrollPos);
                break;
                case KEY_DOWN:
                    l_pViewer->OnVScrollMessage(SB_LINEDOWN, 
                        l_pViewer->m_iScrollPos,
                        l_pViewer->m_iScrollPos);
                break;  
                case KEY_PPAGE:
                    l_pViewer->OnVScrollMessage(SB_PAGEUP, 
                        l_pViewer->m_iScrollPos,
                        l_pViewer->m_iScrollPos);
                break;
                case KEY_NPAGE:
                    l_pViewer->OnVScrollMessage(SB_PAGEDOWN, 
                        l_pViewer->m_iScrollPos,
                        l_pViewer->m_iScrollPos);
                break; 
                case KEY_HOME:case KEY_FIND:
                    l_pViewer->OnVScrollMessage(SB_TOP, 
                        l_pViewer->m_iScrollPos,
                        l_pViewer->m_iScrollPos);
                break; 
                case KEY_END:case KEY_SELECT:
                    l_pViewer->OnVScrollMessage(SB_BOTTOM, 
                        l_pViewer->m_iScrollPos,
                        l_pViewer->m_iScrollPos);
                break;
                case '+': 
                    l_iRefreshDelay += 100;
                    wtimeout(stdscr, l_iRefreshDelay);
                break;
                case '-': 
                    if(l_iRefreshDelay > 100)
                    {
                       l_iRefreshDelay -= 100;
                       wtimeout(stdscr, l_iRefreshDelay);
                    }
                break;
                case 'o':
                    DeleteWindows(l_pViewer);
                    endwin();
                    char * l_pszMessage = MeasureOverhead
                        ("OverheadTrace.txt");  
                    if(!CreateWindows(l_pViewer, 
                        l_iWindowWidth, l_iWindowHeight))
                    {
                        endwin();
                        delete l_pViewer;
                        return FALSE;
                    }
                    mvwprintw(l_pViewer->m_hWnd, 0,0, 
                        l_pszMessage);
                    wprintw(l_pViewer->m_hWnd,"Press any key");
                    wrefresh(l_pViewer->m_hWnd);
                    getch();
                break;               
            }//switch(l_iChar)
            mvwprintw(l_pViewer->m_pWndHelp, 0,0, l_szHelp);   
            wrefresh(l_pViewer->m_pWndHelp);
        }//if(l_iChar != ERR)
        //else
        {
            l_pViewer->OnTimerMessage();
            
            long l_lTime = time(NULL); 
            if(l_pViewer->m_iInvalidated)
            {
                l_pViewer->PaintBuffer(
                    l_pViewer->m_iByteShownFirst, 
                    l_pViewer->m_hWnd, 
                    &l_pViewer->m_ClientRect);
                wrefresh(l_pViewer->m_hWnd);
                l_pViewer->m_iInvalidated = 0;
                l_lTimeRepaint = l_lTime;
            }

            mvwprintw(l_pViewer->m_pWndStatus, 0,0, 
                "Cur Pos %d out of %d. Refresh period %d.",
                l_pViewer->m_iScrollPos, 
                l_pViewer->m_iScrollMax, l_iRefreshDelay);
            wclrtobot(l_pViewer->m_pWndStatus); 
            wrefresh(l_pViewer->m_pWndStatus);                 
        }
    }//while(true)
    DeleteWindows(l_pViewer);
    delete l_pViewer;
    endwin();
    return TRUE;        
}//BOOL ShowDialog()
#endif

BOOL DoWork()
{
    TraceInitialize();
    //By default, show errors and warnings on the
    //debugger's log window and in message boxes.
    TraceAssignGroupsToStream(TO_DebugMonitor | TO_MessageBox,
        TG_MessageBox | TG_Error, (DWORD)-1);

    //Print bank account and thread swaps(if the driver is 
    //loaded) to the memory buffer
    TraceAssignGroupsToStream(TO_MemoryBuffer,
        TG_BankAccount | TG_ThreadSwaps, (DWORD)-1);
    
    //Print application debug messages to the debugger's window
    TraceAssignGroupKeyWordToStream
        (TO_DebugMonitor,KeyWordAppDebug);

    #ifdef WIN32
    GetFullPathName("HTrace.ini", 
        sizeof(s_szFullINIPath), s_szFullINIPath, NULL);

    //Read state of all streams. This can override the 
    //assignments made in the preceding calls.
    TraceReadWriteSettingsFromIniFile(s_szFullINIPath, false,
        (DWORD)-1);
    #endif

    ShowDialog();

    TraceUnInitialize();
    return TRUE;
}//BOOL DoWork()

#if defined(WIN32)
/*------------------------------------------------------------

   FUNCTION: WinMain

   PURPOSE: Initial entry point for the application. 
      
   RETURN VALUE:
      int                        Exit value
------------------------------------------------------------*/
int APIENTRY WinMain
( 
    HINSTANCE p_hInstance, 
    HINSTANCE p_hPrevInstance,
    LPSTR     p_lpszCmdLine, 
    int       p_nCmdShow 
)
{
    DoWork();
    return 1;
}//int APIENTRY WinMain
#else
int main(int argc, char ** argv)
{
    DoWork();
    return 1;
}
#endif

